"""Analytics module protocols."""

from typing import Any, Awaitable, ClassVar, Generic, Protocol, Self, Type, TypeVar, runtime_checkable

import attr

ANALYTICS = str
FRAMERAW = TypeVar("FRAMERAW")
FRAME = TypeVar("FRAME")


class Representable(Protocol):
    """Some container of processing results."""

    def asDict(self):
        """Build results in API format."""


class Error(Representable):
    """Container of processing error."""

    errorCode: int
    description: str
    detail: str


class FrameCtx(Protocol):
    """Frame processing context."""

    def __init__(self, *args, **kwargs): ...

    processed: Representable | dict
    error: Error | None


class FrameRaw(Protocol[FRAMERAW]):
    """Raw frame generic structure"""

    buffer: FRAMERAW
    number: int
    timestamp: float
    lastFrame: bool


@attr.dataclass(slots=True)
class FrameEnviron(Generic[FRAMERAW, FRAME]):
    """Frame environment container."""

    raw: FrameRaw[FRAMERAW]
    frame: FRAME
    result: dict[ANALYTICS, FrameCtx] = attr.field(factory=dict)


class AnalyticsParams(Protocol):
    """Analytics parameters."""

    @classmethod
    def load(cls, obj: Any) -> Self:
        """Load & validate parameters."""


class AnalyticsTargets(Protocol):
    """Analytics targets."""

    @classmethod
    def load(cls, obj: Any) -> Self:
        """Load & validate targets."""


class AnalyticsCtx(Protocol):
    """Analytics context."""

    targets: AnalyticsTargets
    params: AnalyticsParams

    aggregated: Representable | dict
    error: Error | None
    name: str

    def __init__(self, *args, **kwargs): ...

    async def prepareFrameCtx(self, frame: FrameEnviron) -> FrameCtx:
        """
        Prepare frame processing context.

        :param frame: frame loaded by manager

        """


@runtime_checkable
class Node(Protocol):
    """Analytics DAG node protocol."""

    name: str
    requires: ClassVar[list[Type[Self]]]

    async def run(self, frame: FrameCtx, ctx: AnalyticsCtx):
        """
        Frame routine.

        :param frame: frame context
        :param ctx: analytics context

        """


class NodeFollower(Node):
    """Analytics DAG node protocol with optional predictors declared."""

    follows: ClassVar[list[Type[Self]]]


@runtime_checkable
class NodeCollector(Protocol):
    """Analytics DAG result collector node protocol."""

    name: str
    requires: ClassVar[list[Type[Node | Self]]]

    async def collect(self, ctx: AnalyticsCtx):
        """
        Aggregation routine.

        : param ctx: analytics context

        """


@attr.dataclass(slots=True)
class AnalyticsPlan:
    """Analytics manager task."""

    name: str
    ctx: AnalyticsCtx
    targets: set[Type[Node | NodeCollector]]


@runtime_checkable
class Initializable(Protocol):
    """Initializable check object."""

    async def initialize(self, ctx: AnalyticsCtx, *args, **kwargs) -> Awaitable[None | Any]:
        "Initialize node function."
