"""Cache for Analytic Nodes."""

import asyncio
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable, Hashable
from typing import Type


class CacheDict(defaultdict):
    """defaultdict for cache."""

    # pylint: disable=invalid-name
    def setIfNotExist[**P, T](
        self, key: str | Hashable, default: Callable[P, T], *args: P.args, **kwargs: P.kwargs
    ) -> T:
        """Set default if not exists. Return key value.

        if default is an asynchronous function, then an asynchronous task is created, which will then be returned.
        if default is a coroutine, then it wraps into a task that will be returned.
        If default is a normal function, it will be called and its result will be returned.
        If default is not a callable object or a coroutine, its value will simply be returned.

        Args:
            key (str): the key by which the cached value is available
            default: a callable or non-callable object that will be executed / saved to the cache
            args: tuple of arguments for calling the `default` function
            kwargs: dict of arguments for calling the `default` function

        Returns:
            The value stored by key in the cache
        """
        if (value := self.get(key)) is not None:
            return value
        if isinstance(default, Callable):
            if asyncio.iscoroutinefunction(default):
                self[key] = asyncio.create_task(default(*args, **kwargs))
            else:
                self[key] = default(*args, **kwargs)
        elif asyncio.iscoroutine(default):
            self[key] = asyncio.create_task(default)
        else:
            self[key] = default
        return self[key]


class CacheABC(ABC):
    """Cache ABC"""

    _map: defaultdict[Type, defaultdict[int, CacheDict]]

    @classmethod
    @abstractmethod
    def clearCache(cls, nodeType, *args, **kwargs):
        """Clear Cache."""

    @abstractmethod
    def __getitem__(self, item) -> CacheDict: ...

    @abstractmethod
    def __setitem__(self, item, value) -> None: ...

    @abstractmethod
    def clearInstance(self, *args, **kwargs) -> None:
        """Clear current instance cache."""

    @abstractmethod
    def getHash(self, *_, **__) -> int:
        """Instance hash function."""


class CacheDescriptorABC(ABC):
    """Cache Descriptor ABC."""

    def __init__(self, hashFunc: Callable):
        self._hashFunc = hashFunc
        self._cache: CacheABC | None = None
        self._hashArgs: tuple[str, ...] = self._hashFunc.__code__.co_varnames  # hashFunc signature

    @abstractmethod
    def __get__(self, obj, objtype=None): ...

    @abstractmethod
    def cache(self, key: str | Hashable | None = None) -> Callable:
        """Hashable Function Decorator."""


class CacheableType:
    """Type that supports caching."""

    cache: CacheDescriptorABC


class AnalyticsCache[**P, T]:
    """Cache for Analytics.

    The cache is represented by three levels: Class level, Hash level and Cache level.
    * Class level - the class the cache belongs to
    * Hash level - hash of arguments for which the cache is available
    * Cache level - directly dictionary key - value (cache)

    There are two ways to work with the cache:
    1. Create an instance:
        * cache = AnalyticsCache(CacheableType, hashFunc)
        * cache[hashFunc(*args, **kwargs)][key] = 1234
        * cache[hashFunc(*args, **kwargs)].setIfNotExist(key, func, *args, **kwargs)
        * cache[hashFunc(*args, **kwargs)].setIfNotExist(key, data)
        * cache.clearInstance(*args, **kwargs)

    2. or call class methods:
        * AnalyticsCache[ClassType][hashFunc(*args, **kwargs)][key] = 1234
        * AnalyticsCache[ClassType][hashFunc(*args, **kwargs)].setIfNotExist(key, func, *args, **kwargs)
        * AnalyticsCache[ClassType][hashFunc(*args, **kwargs)].setIfNotExist(key, data)
        * AnalyticsCache.clearCache(nodeClass)
        * AnalyticsCache.clearCache(None, *args, **kwargs)
    """

    _map: defaultdict[CacheableType, defaultdict[int, CacheDict]] = defaultdict(
        lambda: defaultdict(lambda: CacheDict(lambda: None))
    )

    @classmethod
    # pylint: disable=keyword-arg-before-vararg
    def clearCache(cls, nodeType: CacheableType | None = None, *args, **kwargs) -> None:
        """Clear Cache by params.

        Args:
            nodeType: if specified, the cache will be cleared only for this node class
            args: if specified, the cache will be cleared only for node run argumments

        Examples:
            * clearCache(HumanTracking, (frameCtx, ctx)) - delete cache where class HumanTracking and run args:
            frameCtx, ctx
            * clearCache() - delete all cache
            * clearCache(HumanTracking) - delete all cache of HumanTracking class
            * clearCache(args=(frameCtx, ctx)) - remove from all classes the cache containing runs with arguments:
            frameCtx, ctx
        """
        if not nodeType and not args and not kwargs:
            cls._map.clear()
        elif nodeType and (args or kwargs):
            cls._map[nodeType].pop(nodeType.cache.hashFunc(*args, **kwargs), None)
        elif nodeType:
            cls._map[nodeType].clear()
        elif args or kwargs:
            for classType, cache in cls._map.items():
                try:
                    cache.pop(classType.cache.hashFunc(*args, **kwargs), None)
                except TypeError:
                    ...

    def __init__(self, nodeClass: CacheableType, hashFunc: Callable[P, int]):
        self.nodeClass = nodeClass
        self.hashFunc = hashFunc

    def __getitem__(self, item):
        return self._map[self.nodeClass][item]

    def __setitem__(self, item, value):
        self._map[self.nodeClass][item] = value

    def clearInstance(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Clear cash for current instance.

        Args, kwargs: arguments for hash function.

        """
        if not args and not kwargs:
            self._map[self.nodeClass].clear()
        else:
            self._map[self.nodeClass].pop(self.getHash(*args, **kwargs), None)

    def getHash(self, *args: P.args, **kwargs: P.kwargs) -> int:
        """Proxy for hashFunc."""
        return self.hashFunc(*args, **kwargs)


def getSignature(func) -> tuple[str, ...]:
    """Function Attributes Signature."""
    return func.__code__.co_varnames[: func.__code__.co_argcount]


class CacheDecorator[**P, T]:
    """Cache Decorator.

    Attributes:
        func: wrapped function
        key: the key by which the cached value will be available
        hashFunc: hash function
        _cache: cache instance, created at the time of initialization of the class function
        _hashArgs: hash function arguments signature
        _funcArgs: wrapped function arguments signature
        _bound: bound method of wrapped function. Created when a function is called with an instance of its class
    """

    def __init__(self, func: Callable[P, T], key, hashFunc: Callable):
        self.func = func
        self.key = key
        self.hashFunc = hashFunc

        self._cache: AnalyticsCache | None = None
        self._hashArgs: tuple[str, ...] = getSignature(hashFunc)
        self._funcArgs: tuple[str, ...] = getSignature(func)[1:]  # remove self; its only bound method
        self._bound: Callable[P, T] | None = None

    def __get__(self, obj, type_=None) -> Callable[P, T]:  # pylint: disable=invalid-name
        self._bound = self.func.__get__(obj, type_)  # wrap func as class method
        if not self._cache:
            self._cache = AnalyticsCache(nodeClass=type_, hashFunc=self.hashFunc)
        return self.__call__

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        _kwargs = kwargs.copy()
        _kwargs.update(zip(self._funcArgs, args))
        localCache = self._cache[self.hashFunc(**{arg: _kwargs[arg] for arg in self._hashArgs})]  # type: ignore
        result = localCache.setIfNotExist(self.key, self._bound, *args, **kwargs)
        return result


class CacheDescriptor[**P, T]:
    """Cache Descriptor."""

    def __init__(self, hashFunc: Callable[P, T]):
        self._hashFunc = hashFunc
        self._cache: AnalyticsCache | None = None

    def __get__(self, obj, objtype=None):
        if not self._cache:
            self._cache = AnalyticsCache(nodeClass=objtype, hashFunc=self._hashFunc)
        return self._cache

    def cache(self, key: str | Hashable):
        """Cache Decorator.

        Args:
            key: any hashable value. The key by which to save the result.
        """

        def decorator[**S, R](func: Callable[S, R]) -> Callable[S, R]:  # pylint: disable=invalid-name
            return CacheDecorator(func, key, self._hashFunc)

        return decorator
