"""Analytics module loader."""

import importlib
from collections.abc import Iterable
from typing import Any, Protocol, Type, runtime_checkable

from pydantic import ValidationError

from luna_analytics_manager.base import (
    AnalyticsCtx,
    AnalyticsParams,
    AnalyticsPlan,
    AnalyticsTargets,
    FrameCtx,
    Node,
    NodeCollector,
)
from luna_analytics_manager.errors import AnalyticsLoaderException


@runtime_checkable
class AnalyticsModule(Protocol):
    """
    Analytics module protocol.

    """

    __name__: str

    #: Container for user-defined estimation targets to be loaded in
    Targets: Type[AnalyticsTargets]
    #: Container for user-defined configurable parameters to be loaded in
    Params: Type[AnalyticsParams]
    #: Analytics context to share overall processing results & presets between different nodes run on different frames
    AnalyticsCtx: Type[AnalyticsCtx]
    #: Frame context to share certain frame processing results between different nodes run on this frame
    FrameCtx: Type[FrameCtx]

    async def initialize(self, **kwargs):
        """
        Pre-initializing routine.
        One-time load of external dependencies (e.g. database tools, neural network models) should be done here.
        """

    def targetNodes(self, targets: AnalyticsTargets) -> Iterable[Type[Node | NodeCollector]]:
        """
        Pre-select nodes by user-defined targets.

        :param targets: user-defined targets loaded

        """

    def mandatoryNodes(
        self, params: AnalyticsParams, targets: AnalyticsTargets
    ) -> Iterable[Type[Node | NodeCollector]]:
        """
        Pre-select required nodes according to configuration parameters.

        :param params: user-defined configuration parameters loaded
        :param targets: user-defined targets loaded

        """


async def preload(name: str, **kwargs) -> AnalyticsModule:
    """
    Load & pre-initialize analytics by module name.

    :param name: module name specified in absolute terms to be passed in
      `importlib.import_module <https://docs.python.org/3/library/importlib.html#importlib.import_module>`_
    :raises AnalyticsLoaderException: module not found, or not comply with protocol

    """

    try:
        module = importlib.import_module(name)
    except ModuleNotFoundError as exc:
        raise AnalyticsLoaderException(f"Module {name} not found") from exc

    if not isinstance(module, AnalyticsModule):
        raise AnalyticsLoaderException(f"Module {name} does not comply with protocol")

    await module.initialize(**kwargs)
    return module


def planFactory(
    analytics: AnalyticsModule,
    name: str = None,
    *,
    targets: Any = None,
    settings: Any = None,
    **contextKwargs,
) -> AnalyticsPlan:
    """
    Analytics plan factory.

    :param analytics: analytics module
    :param name: plan name (omit to use module name)
    :param targets: user-defined targets to load :func:`~luna_analytics_manager.base.AnalyticsTargets`
    :param settings: user-defined parameters to load :func:`~luna_analytics_manager.base.AnalyticsParams`
    :param contextKwargs: context keyword arguments
    :raises AnalyticsLoaderException: failed to load analytics targets, or parameters

    """

    name = name or analytics.__name__.split(".")[-1]

    try:
        targets = analytics.Targets.load(targets)
    except ValidationError as exc:
        error = exc.errors()[0]
        msg = f"Path: '{'.'.join(map(str, error['loc']))}', message: '{error['msg']}'"
        raise AnalyticsLoaderException(f"Failed to load '{name}' analytics targets. {msg}") from exc
    except Exception as exc:
        raise AnalyticsLoaderException(f"Failed to load '{name}' analytics targets") from exc

    try:
        params = analytics.Params.load(settings)
    except ValidationError as exc:
        error = exc.errors()[0]
        msg = f"Path: '{'.'.join(map(str, error['loc']))}', message: '{error['msg']}'"
        raise AnalyticsLoaderException(f"Failed to load '{name}' analytics parameters. {msg}") from exc
    except Exception as exc:
        raise AnalyticsLoaderException(f"Failed to load '{name}' analytics parameters") from exc

    try:
        context = analytics.AnalyticsCtx(params=params, targets=targets, **contextKwargs)
    except ValidationError as exc:
        error = exc.errors()[0]
        raise AnalyticsLoaderException(f"Failed to load '{name}' analytic. {error['msg']}") from exc
    targets = set(analytics.targetNodes(targets)) | set(analytics.mandatoryNodes(params, targets))
    plan = AnalyticsPlan(name=name, ctx=context, targets=targets)
    return plan
