"""Analytics manager implementation."""

import asyncio
from asyncio import Event, Queue, Task, TaskGroup
from collections.abc import AsyncIterable
from contextlib import asynccontextmanager
from functools import partial
from itertools import chain
from logging import Logger
from time import perf_counter
from typing import Any, AsyncGenerator, AsyncIterator, Callable, Coroutine, Generic, Iterable, Type, cast

from .base import (
    FRAME,
    FRAMERAW,
    AnalyticsCtx,
    AnalyticsPlan,
    FrameEnviron,
    FrameRaw,
    Initializable,
    Node,
    NodeCollector,
)
from .cache import AnalyticsCache
from .errors import (
    AnalyticsException,
    AnalyticsFailure,
    FrameCtxFailure,
    FrameLoadFailure,
    ManagerPublicException,
    VideoException,
)
from .graph import Graph
from .monitoring.monitoring import FrameMonitoringPoint
from .monitoring.storage import FRAMES_STORAGE
from .rate import FrameRateChecker
from .sync import Syncer

ANALYTICS = str
NODE = Node | NodeCollector

logger = Logger("luna.analytics_manager")


# pylint: disable=too-many-instance-attributes,too-many-locals
class AnalyticsManager(Generic[FRAMERAW, FRAME]):
    """
    DAG analytics manager.
    Attributes:
        contexts: map of analytics module name to predefined analytics context
        collection: map of analytics module name to bounden analytics nodes
        _frameGraphs: map of analytics module name to frame DAG internal data
        _collectorGraphs: map of analytics module name to collector DAG internal data
        syncer: node processing sync util to call the certain node in frame-by-frame order
        rater: frame processing rate checker
        concurrency: limit on the number of frames processed in parallel
        _waitingForProcess: whether the processing not run yet

    """

    #: Load frame in format acceptable for all the analytics
    loadFrame: Callable[[FRAMERAW], FRAME]
    #: Get frame processing rate checker for certain analytics
    getRater: Callable[[AnalyticsCtx], FrameRateChecker]

    def __init__(
        self,
        plans: Iterable[AnalyticsPlan],
        concurrency: int = 10,
        additionalNodeData: dict[type[Node] | type[NodeCollector], dict[str, Any]] | None = None,
    ):
        """
        :param plans: analytics plans configured
        :param concurrency: frame processing concurrency
        """

        self.contexts: dict[ANALYTICS, AnalyticsCtx] = {}
        self.collection: dict[ANALYTICS, dict[Type[NODE], NODE]] = {}
        self._frameGraphs: dict[ANALYTICS, dict] = {}
        self._collectorGraphs: dict[ANALYTICS, dict] = {}
        for plan in plans:
            analytics = plan.name
            if analytics in self.contexts:
                raise RuntimeError(f"plan with name {analytics} already exists")
            self.contexts[analytics] = plan.ctx
            getRequires: Callable = lambda n: n.requires + [*chain(*map(getRequires, n.requires))]
            nodesToRun = tuple(chain(plan.targets, *map(getRequires, plan.targets)))
            self.collection[analytics] = {node: node() for node in nodesToRun}
            getParents: Callable = lambda n, allNodes: n.requires + [*set(getattr(n, "follows", [])) & set(allNodes)]
            self._frameGraphs[analytics] = {
                node: {parent for parent in getParents(node, nodesToRun) if isinstance(parent, Node)}
                for node in nodesToRun
                if isinstance(node, Node)
            }
            self._collectorGraphs[analytics] = {
                node: {parent for parent in getParents(node, nodesToRun) if isinstance(parent, NodeCollector)}
                for node in nodesToRun
                if isinstance(node, NodeCollector)
            }
        self.syncer: dict[ANALYTICS, Syncer[Type[NODE]]] = {analytics: Syncer() for analytics in self.collection}
        self.rater: dict[ANALYTICS, FrameRateChecker] = {
            analytics: self.getRater(ctx) for analytics, ctx in self.contexts.items()
        }
        self.additionalNodeData = additionalNodeData or {}
        self.concurrency = concurrency
        self._waitingForProcess: bool = True
        self._scheduledToStop: bool = False
        self._stopEvent: asyncio.Future | None = None

    def frameGraph(self, analytics: ANALYTICS) -> Graph[Type[Node]]:
        """Instantiate a frame analytics graph."""
        graph = Graph[Type[Node]](self._frameGraphs[analytics])
        graph.prepare()
        return graph

    def collectorGraph(self, analytics: ANALYTICS) -> Graph[Type[NodeCollector]]:
        """Instantiate an analytics collector graph."""
        graph = Graph[Type[NodeCollector]](self._collectorGraphs[analytics])
        graph.prepare()
        return graph

    @staticmethod
    def _terminate(eg: ExceptionGroup, tasks: list[Task]):
        """Catch the process interruption details, and raise particular exception to terminate processing."""

        sources = [task.get_name() for task in tasks if not task.cancelled() and task.exception()]
        msg = f"Process interrupted due to {len(eg.exceptions)} unhandled error(s) in {','.join(sources)}"
        raise AnalyticsFailure(msg) from eg

    async def _execute(self, job: Coroutine):
        """Wrap node job to perform some monitoring actions."""

        await job
        logger.debug("Node job %s completed", job.__qualname__)

    async def _processAll(self, analytics: ANALYTICS):
        """Assign tasks for aggregation along DAG, and wait for results."""

        graph = self.collectorGraph(analytics)
        ctx = self.contexts[analytics]

        tasks = []
        try:
            async with TaskGroup() as collectorTasks:
                async for target in graph.nodes():
                    if ctx.error:
                        break
                    node = cast(NodeCollector, self.collection[analytics][target])
                    task = collectorTasks.create_task(self._execute(node.collect(ctx)))
                    task.set_name(f"{analytics}.{target.__name__}.collect")
                    task.add_done_callback(partial((lambda g, n, _: g.done(n)), graph, target))
                    tasks.append(task)
        except* AnalyticsException:
            raise
        except* Exception as eg:  # pylint: disable=broad-exception-caught
            self._terminate(eg, tasks)

        logger.debug("Analytics %s collect completed", analytics)

    async def _collectAll(self):
        """Assign tasks for aggregation along DAG with predefined analytics, and wait for results."""

        async with TaskGroup() as analyticsTasks:
            for analytics in self.collection:
                analyticsTasks.create_task(self._processAll(analytics))

    async def _processOne(self, frame: FrameEnviron, analytics: ANALYTICS, point: FrameMonitoringPoint):
        """Prepare frame context, assign tasks for processing a single frame along DAG, and wait for results ."""

        start = perf_counter()
        graph = self.frameGraph(analytics)
        ctx = self.contexts[analytics]
        try:
            frameCtx = frame.result[analytics] = await ctx.prepareFrameCtx(frame)
        except Exception as exc:
            raise FrameCtxFailure(f"Failed to load context for frame number {frame.raw.number}") from exc

        tasks = []
        try:
            async with self.syncer[analytics].process(frame.raw.number), TaskGroup() as frameTasks:
                isSkipping = self.rater[analytics].checkFrameSkip(frameCtx)
                async for target in graph.nodes():
                    async with self.syncer[analytics].run(frame.raw.number, target):
                        if isSkipping or frameCtx.error or ctx.error:
                            graph.done(target)
                            continue
                        node = cast(Node, self.collection[analytics][target])
                        task = frameTasks.create_task(self._execute(node.run(frameCtx, ctx)))
                        task.set_name(f"{analytics}.{target.__name__}.run for frame number {frame.raw.number}")
                        task.add_done_callback(partial((lambda g, n, _: g.done(n)), graph, target))
                        tasks.append(task)
        except* AnalyticsException:
            raise
        except* Exception as eg:  # pylint: disable=broad-exception-caught
            self._terminate(eg, tasks)
        finally:
            point.analyticsExecutionTime[ctx.name + "_frame_execution"] = perf_counter() - start
            AnalyticsCache.clearCache(None, frameCtx, ctx)

        logger.debug("Analytics %s run completed for frame number %i", analytics, frame.raw.number)

    async def _runOne(self, frame: FrameEnviron, precessor: Event):
        """Run task for processing a single frame with predefined analytics, and wait for previous frame get ready."""

        with FRAMES_STORAGE.monitoring(FrameMonitoringPoint(self.concurrency)) as point:
            async with TaskGroup() as analyticsTasks:
                for analytics in self.collection:
                    analyticsTasks.create_task(self._processOne(frame, analytics, point))
            await precessor.wait()

    async def _getFrame(self, rawframe: FrameRaw[FRAMERAW]) -> FrameEnviron:
        """Prepare frame environment container."""

        try:
            frame: FrameEnviron[FRAMERAW, FRAME] = FrameEnviron(rawframe, self.loadFrame(rawframe.buffer))
        except Exception as exc:
            raise FrameLoadFailure(f"Failed to load frame number {rawframe.number}") from exc
        return frame

    async def _run(self, frames: AsyncIterable[FrameRaw[FRAMERAW]], done: Queue, readyForNext: Event):
        """Concurrently process a sequence of frames, and put frame results to queue."""

        async with TaskGroup() as frameTasks:
            (lastFrameEvent := Event()).set()
            pending: dict[Task, FrameEnviron] = {}
            frameIter = aiter(frames)
            while True:
                try:
                    frame = await anext(frameIter)
                except StopAsyncIteration:
                    break
                except Exception as e:
                    raise VideoException(videoException=e) from e
                frameEvent = Event()
                frameEnv = await self._getFrame(frame)
                task = frameTasks.create_task(self._runOne(frameEnv, lastFrameEvent))
                task.add_done_callback(partial(lambda e, t: [done.put_nowait(pending.pop(t)), e.set()], frameEvent))
                pending[task] = frameEnv
                lastFrameEvent = frameEvent
                if len(pending) + done.qsize() >= self.concurrency:
                    readyForNext.clear()
                    await readyForNext.wait()
        await self._collectAll()

    @asynccontextmanager
    async def process(
        self, frames: AsyncIterable[FrameRaw[FRAMERAW]]
    ) -> AsyncIterator[AsyncGenerator[AsyncIterable[FrameRaw[FRAMERAW]], FrameEnviron]]:
        """Control async generator finalization"""
        process = self._process(frames)
        try:
            yield process
        finally:
            if process.ag_running:  # type: ignore # pylint: disable=using-constant-test
                await self._scheduleToStop()
            await process.aclose()

    async def _scheduleToStop(self):
        """Schedule process generator to stop iterating"""
        self._scheduledToStop = True
        self._stopEvent = asyncio.get_event_loop().create_future()
        await self._stopEvent

    async def _process(
        self, frames: AsyncIterable[FrameRaw[FRAMERAW]]
    ) -> AsyncGenerator[AsyncIterable[FrameRaw[FRAMERAW]], FrameEnviron]:
        """
        Create task for processing, and wait for frame results from queue.

        :param frames: frames to run analytics for
        :raises ManagerPublicException: processing interrupted due to analytics errors

        """

        assert self._waitingForProcess, "Process can only be run once"
        self._waitingForProcess = False

        stopword = "COMPLETE"

        async with asyncio.TaskGroup() as tg:
            for analytics, nodesDict in self.collection.items():
                ctx = self.contexts[analytics]
                for nodeType, nodeObj in nodesDict.items():
                    if isinstance(nodeObj, Initializable):
                        tg.create_task(nodeObj.initialize(ctx, **self.additionalNodeData.get(nodeType, {})))

        wasException = False  # necessary to prevent group exceptions
        try:
            async with TaskGroup() as tg:
                readyForNext = Event()
                doneQueue: Queue = Queue()
                runTask = tg.create_task(self._run(frames, doneQueue, readyForNext))
                runTask.add_done_callback(lambda _: doneQueue.put_nowait(stopword))

                while (result := await doneQueue.get()) != stopword:
                    yield result
                    if self._scheduledToStop:
                        raise GeneratorExit
                    readyForNext.set()
        except* VideoException as eg:
            if not wasException:
                wasException = True
                # catch decoding error from self._run
                if isinstance(eg, VideoException):
                    raise
                raise eg.exceptions[0].exceptions[0].originalException  # type: ignore # pylint: disable=no-member
            logger.warning("a higher-level exception was raised, the current one (VideoException) will be ignored.")
        except* AnalyticsException as eg:
            if not wasException:
                wasException = True
                expand: Callable = lambda e: (
                    [*chain(*map(expand, e.exceptions))] if isinstance(e, ExceptionGroup) else [e]  # type: ignore
                )
                errors = expand(eg)
                msg = f"Some errors occurred while processing: {[e.args[0] for e in errors]}"
                raise ManagerPublicException(msg, errors=errors) from eg
            logger.warning("a higher-level exception was raised, the current one (AnalyticsException) will be ignored.")
        except* GeneratorExit as eg:
            if not wasException:
                wasException = True
                raise eg.exceptions[0]  # type: ignore # pylint: disable=no-member
        finally:
            if self._stopEvent:
                self._stopEvent.set_result(True)
