"""Module contains global monitoring pont storages"""

from abc import abstractmethod
from contextvars import ContextVar
from datetime import datetime

videoIdCtx: ContextVar[str] = ContextVar("videoId", default="")


class BaseMonitoringPoint:
    """Analytics base monitoring point."""

    name: str

    __slots__ = ("videoId", "eventTime", "executionTime")

    def __init__(self, executionTime: float = None, eventTime: float = None):
        self.videoId = videoIdCtx.get()
        self.eventTime = eventTime or datetime.now().timestamp()
        self.executionTime = executionTime

    def getFields(self) -> dict:
        """Get monitoring fields"""
        return {"execution_time": self.executionTime, "video_id": self.videoId}

    @abstractmethod
    def getTags(self) -> dict:
        """Get monitoring tags"""
        raise NotImplementedError


class FrameMonitoringPoint(BaseMonitoringPoint):
    """Base frame processing monitoring point"""

    name = "frame_monitoring"

    def __init__(self, concurrency: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.concurrency = concurrency
        self.analyticsExecutionTime: dict[str, float] = {}

    def getFields(self) -> dict:
        """Get monitoring fields"""
        fields = super().getFields()
        fields.update(self.analyticsExecutionTime)
        return fields

    def getTags(self) -> dict:
        """Get monitoring tags"""
        return {"concurrency": self.concurrency}


class NodeMonitoringPoint(BaseMonitoringPoint):
    """Node execution monitoring point"""

    __slots__ = ("skip", "name")

    def __init__(self, name: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.skip = False
        self.name = name

    def getTags(self) -> dict:
        """Get monitoring tags"""
        return {"skipped": str(self.skip).lower()}
