"""Frame processing sync util."""

import asyncio
from contextlib import asynccontextmanager
from contextvars import ContextVar
from typing import Generic, TypeVar

NODE = TypeVar("NODE")
frameIdCtx: ContextVar[int] = ContextVar("frameId")


class Syncer(Generic[NODE]):
    """
    Frame processing sync util to run the certain handler in frame-by-frame order.
    At the start of processing a new frame it is registered for the sync.
    At the end of processing a frame, the previous frame is removed from the sync cause it no longer has any waiters.

    Single-node synchronization example:

        >>> import asyncio
        >>> import random

        >>> syncer = Syncer()

        >>> async def process_frame(frame_number):
        ...
        ...     # it is guaranteed the order of result output corresponds to the frame order
        ...     async with syncer.process(frame_number, singlenode=True):
        ...         await asyncio.sleep(random.randint(1, 10)/10)
        ...         print(frame_number)

        >>> async def main():
        ...     await asyncio.gather(*map(process_frame, [1, 2]))

        >>> asyncio.run(main())
        1
        2

    Multi-node synchronization example:

        >>> async def process_frame(frame_number):
        ...
        ...     async with syncer.process(frame_number):
        ...         for node in [...]:
        ...
        ...             # it is guaranteed the order of result output corresponds to the frame order
        ...             async with syncer.run(frame_number, node):
        ...                 await asyncio.sleep(random.randint(1, 10)/10)
        ...                 print(frame_number)

        >>> async def main():
        ...     await asyncio.gather(*map(process_frame, [1, 2]))

        >>> asyncio.run(main())
        1
        2

    """

    def __init__(self):
        self.frames = {}

    def _add(self, frameId: int):
        """Add frame if no waiter has added it yet."""

        if frameId not in self.frames:
            self.frames[frameId] = {}

    def _remove(self, frameId: int):
        """Remove frame."""

        self.frames.pop(frameId, None)

    @asynccontextmanager
    async def process(self, frameId, singlenode: bool = False):
        """Process frame."""

        try:
            self._add(frameId)
            if singlenode:
                async with self.run(frameId, ...):  # type: ignore
                    yield
            else:
                yield
        finally:
            self._remove(frameId - 1)

    async def _wait(self, frameId: int, node: NODE):
        """Wait for node called. If this is the first waiter for this frame node, create an event at once."""

        if node not in self.frames.setdefault(frameId, {}):
            self.frames[frameId][node] = asyncio.Event()
        await self.frames[frameId][node].wait()

    def _set(self, frameId: int, node: NODE):
        """Mark node as called. If no one is waiting for this frame node yet, create an event that set at once."""

        if node not in self.frames.setdefault(frameId, {}):
            self.frames[frameId][node] = asyncio.Event()
        self.frames[frameId][node].set()

    @asynccontextmanager
    async def run(self, frameId: int, node: NODE):
        """Run node."""

        try:
            if frameId > 1:
                await self._wait(frameId - 1, node)
            yield
        finally:
            self._set(frameId, node)


class Syncer2:
    """
    Frame processing sync util to run the certain handler in frame-by-frame order.
    At the start of processing a new frame it is registered for the sync.
    To run frame processing scope, it is waited for no frame with a lower number is currently processed.
    At the end of processing a frame, it is removed from the sync, and all the waiters are notified.

    """

    def __init__(self):
        self.frames = {}

    @asynccontextmanager
    async def process(self, frameId):
        """Register frame for processing."""

        self._register(frameId)
        try:
            yield self
        finally:
            self._unregister(frameId)

    def _register(self, frameId):
        """Add frame."""

        frameIdCtx.set(frameId)
        self.frames[frameId] = asyncio.Event()

    def _unregister(self, frameId):
        """Mark frame as processed."""

        self.frames.pop(frameId).set()

    async def _wait(self):
        """Wait for all the frame predecessors processed."""

        currentFrameId = frameIdCtx.get()
        while frameToWait := next(filter(lambda fid: fid < currentFrameId, self.frames), None):
            await self.frames[frameToWait].wait()

    @asynccontextmanager
    async def acquire(self):
        """Run processing scope."""

        await self._wait()
        yield
