"""Human tracking analytics aggregation."""

# mypy: disable-error-code="union-attr"

import asyncio
from functools import partial
from typing import Literal

from lunavl.sdk.descriptors.descriptors import BodyDescriptor, FaceDescriptor
from lunavl.sdk.detectors.bodydetector import BodyDetection
from lunavl.sdk.detectors.facedetector import FaceDetection
from lunavl.sdk.estimators.body_estimators.body_attributes import BodyAttributes
from lunavl.sdk.estimators.face_estimators.basic_attributes import BasicAttributes
from lunavl.sdk.estimators.face_estimators.deepfake import Deepfake
from lunavl.sdk.estimators.face_estimators.livenessv1 import LivenessV1
from lunavl.sdk.image_utils.geometry import Rect
from sdk_loop.estimators.body_estimators.body_attributes_estimator import BodyAttributesEstimationBroker
from sdk_loop.estimators.body_estimators.body_descriptor_estimator import AggregatedBodyDescriptorEstimationBroker
from sdk_loop.estimators.face_estimators.basic_attributes_estimator import AggregatedBasicAttributesEstimationBroker
from sdk_loop.estimators.face_estimators.deepfake_estimator import (
    DeepfakeInputData,
    DeepfakeQueueEstimationBrokerBase,
    aggregateDeepfakeFromEstimates,
)
from sdk_loop.estimators.face_estimators.face_descriptor_estimator import AggregatedFaceDescriptorEstimationBroker
from sdk_loop.estimators.face_estimators.liveness1_estimator import (
    LivenessInputData,
    Livenessv1QueueEstimationBroker,
    aggregateLivenessFromEstimates,
)

from luna_analytics_manager.monitoring.node_monitoring import nodeMonitoring
from luna_analytics_manager.sync import Syncer2 as Syncer
from luna_analytics_manager.templates.image_helpers import toBytes
from luna_analytics_manager.templates.models import EventStatus, ImageRetainPolicy

from .analytics import AnalyticsCtx
from .estimations import NodeBodyWarp, NodeFaceWarp, NodeLandmarks5
from .event_policy import NodeHumanTrackingEventPolicy
from .frame_structures import (
    AttributesInfo,
    BodyEstimatedAttributes,
    FaceEstimatedAttributes,
    FrameCtx,
    HumanTrack,
    ShotInfo,
)


class NodeAggregation:
    """Extra aggregation node."""

    name: str

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Perform estimations."""

    @nodeMonitoring()
    async def run(self, frameCtx: FrameCtx, ctx: AnalyticsCtx):
        """Run estimation on frame tracks."""

        if frameCtx.processed is None or frameCtx.processed.humanTracks is None:
            return

        await self._run(frameCtx.processed.humanTracks, ctx)


def getEventsAndShots(humanTracks: list, shotsType: Literal["body", "face"]) -> tuple[list, list]:
    """Constructs events and ones shots lists"""
    events = []
    shots = []

    for track in humanTracks:
        if not (event := track.currentEvent) or not event.aggregated:
            continue
        events.append(event)
        eventShots = event.eventShots.faceShots if shotsType == "face" else event.eventShots.bodyShots
        shots.append(eventShots)

    return events, shots


class NodeLivenessV1(NodeAggregation):
    """Liveness estimation node."""

    name = "aggregated_liveness"
    requires = [NodeLandmarks5, NodeHumanTrackingEventPolicy]

    async def estimate(
        self, shots: list[AttributesInfo[FaceEstimatedAttributes]], estimator: Livenessv1QueueEstimationBroker
    ) -> list[LivenessV1] | None:
        if not shots:
            return None

        coros = (estimator.estimate(LivenessInputData(shot.attributes.detection)) for shot in shots)
        attributes = await asyncio.gather(*coros)
        return attributes

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Run liveness estimation on the finalized events."""

        threshold = ctx.sdk.livenessV1Settings.scoreThreshold
        estimator = ctx.sdk.engine.estimators.getLivenessv1Estimator()

        events, shots = getEventsAndShots(humanTracks, "face")
        results = await asyncio.gather(*map(partial(self.estimate, estimator=estimator), shots))
        for event, attributes in zip(events, results):
            liveness = aggregateLivenessFromEstimates(attributes, threshold) if attributes else None
            event.aggregated.aggregatedEstimations.face.liveness = liveness


class NodeDeepfake(NodeAggregation):
    """Deepfake estimation node."""

    name = "aggregated_deepfake"
    requires = [NodeHumanTrackingEventPolicy]

    async def estimate(
        self, shots: list[AttributesInfo[FaceEstimatedAttributes]], estimator: DeepfakeQueueEstimationBrokerBase
    ) -> list[Deepfake] | None:
        if not shots:
            return None

        coros = (estimator.estimate(DeepfakeInputData(shot.attributes.detection)) for shot in shots)
        attributes = await asyncio.gather(*coros)
        return attributes

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Run deepfake estimation on the finalized events."""

        config = ctx.sdk.deepfakeSettings
        match config.mode.value:
            case 1:
                estimator = ctx.sdk.engine.estimators.getDeepfakeV1Estimator()
            case 2:
                estimator = ctx.sdk.engine.estimators.getDeepfakeV2Estimator()
            case _:
                raise RuntimeError("Unknown configuration")

        events, shots = getEventsAndShots(humanTracks, "face")
        results = await asyncio.gather(*map(partial(self.estimate, estimator=estimator), shots))
        for event, attributes in zip(events, results):
            deepfake = aggregateDeepfakeFromEstimates(attributes, config.scoreThreshold) if attributes else None
            event.aggregated.aggregatedEstimations.face.deepfake = deepfake


class NodeFaceShots(NodeAggregation):
    """
    Face shots node.
    Attributes:
        config: shot retain config, prevents quality loss since it can be used for re-extraction
    """

    name = "aggregated_face_samples"
    requires = [NodeFaceWarp, NodeHumanTrackingEventPolicy]

    def __init__(self):
        self.config = ImageRetainPolicy(mimetype="JPEG", quality=1)

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Run face best shot selection on the finalized events."""

        for track in humanTracks:
            if not (event := track.currentEvent):
                continue
            if event.aggregated:
                attributes = event.eventShots.faceShots
                images = (attrInfo.attributes.warp.warp.warpedImage for attrInfo in attributes)
                warps = await asyncio.gather(*map(partial(toBytes, config=self.config), images))
                if warps:
                    event.aggregated.aggregatedEstimations.face.shots = [
                        ShotInfo(shot.timeOffset, warp) for shot, (warp, _) in zip(attributes, warps)
                    ]


class NodeFaceBestshot(NodeAggregation):
    """Face best shot node."""

    name = "aggregated_face_thumbnail"
    requires = [NodeFaceWarp, NodeHumanTrackingEventPolicy]

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Run face best shot selection on the finalized events."""

        for track in humanTracks:
            if not (event := track.currentEvent):
                continue
            if event.aggregated:
                if shots := event.eventShots.faceShots:
                    warp, _ = await toBytes(shots[0].attributes.warp.warp.warpedImage, ctx.params.imageRetainPolicy)
                    bestshot = ShotInfo(shots[0].timeOffset, warp)
                    event.aggregated.aggregatedEstimations.face.bestshot = bestshot


class NodeFaceBasicAttributes(NodeAggregation):
    """Face basic attributes estimation node."""

    name = "aggregated_face_basic_attributes"
    requires = [NodeFaceWarp, NodeHumanTrackingEventPolicy]

    async def estimate(
        self, shots: list[AttributesInfo[FaceEstimatedAttributes]], estimator: AggregatedBasicAttributesEstimationBroker
    ) -> BasicAttributes | None:
        if not shots:
            return None

        warps = [shot.attributes.warp.warp for shot in shots]
        _, attributes = await estimator.estimate(warps)
        return attributes

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Run face basic attributes estimation on the finalized events."""

        estimator = ctx.sdk.engine.estimators.getAggreagtedBasicAttributesEstimator()

        events, shots = getEventsAndShots(humanTracks, "face")
        results = await asyncio.gather(*map(partial(self.estimate, estimator=estimator), shots))
        for event, attributes in zip(events, results):
            event.aggregated.aggregatedEstimations.face.basicAttributes = attributes


class NodeFaceDescriptor(NodeAggregation):
    """Face descriptor estimation node."""

    name = "aggregated_face_descriptor"
    requires = [NodeLandmarks5, NodeFaceWarp, NodeHumanTrackingEventPolicy]

    async def estimate(
        self, shots: list[AttributesInfo[FaceEstimatedAttributes]], estimator: AggregatedFaceDescriptorEstimationBroker
    ) -> FaceDescriptor | None:
        if not shots:
            return None

        warps = [shot.attributes.warp.warp for shot in shots]
        _, descriptor = await estimator.estimate(warps)
        return descriptor

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Run face descriptor estimation on finalised event."""

        estimator = ctx.sdk.engine.estimators.getAggreagtedFaceDescriptorEstimator()

        events, shots = getEventsAndShots(humanTracks, "face")
        results = await asyncio.gather(*map(partial(self.estimate, estimator=estimator), shots))
        for event, descriptor in zip(events, results):
            event.aggregated.aggregatedEstimations.face.descriptor = descriptor


class NodeBodyShots(NodeAggregation):
    """
    Body shot node.
    Attributes:
        config: shot retain config, prevents quality loss since it can be used for re-extraction
    """

    name = "aggregated_body_samples"
    requires = [NodeBodyWarp, NodeHumanTrackingEventPolicy]

    def __init__(self):
        self.config = ImageRetainPolicy(mimetype="JPEG", quality=1)

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Run face best shot selection on the finalized events."""

        for track in humanTracks:
            if not (event := track.currentEvent):
                continue
            if event.aggregated:
                attributes = event.eventShots.bodyShots
                images = (attrInfo.attributes.warp.warpedImage for attrInfo in attributes)
                warps = await asyncio.gather(*map(partial(toBytes, config=self.config), images))
                if warps:
                    event.aggregated.aggregatedEstimations.body.shots = [
                        ShotInfo(shot.timeOffset, warp) for shot, (warp, _) in zip(attributes, warps)
                    ]


class NodeBodyBestshot(NodeAggregation):
    """Body best shot node."""

    name = "aggregated_body_thumbnail"
    requires = [NodeBodyWarp, NodeHumanTrackingEventPolicy]

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Run face best shot selection on the finalized events."""

        for track in humanTracks:
            if not (event := track.currentEvent):
                continue
            if event.aggregated:
                if shots := event.eventShots.bodyShots:
                    warp, _ = await toBytes(shots[0].attributes.warp.warpedImage, ctx.params.imageRetainPolicy)
                    bestshot = ShotInfo(shots[0].timeOffset, warp)
                    event.aggregated.aggregatedEstimations.body.bestshot = bestshot


class NodeBodyDescriptor(NodeAggregation):
    """Face descriptor estimation node."""

    name = "aggregated_body_descriptor"
    requires = [
        NodeBodyWarp,
        NodeHumanTrackingEventPolicy,
    ]

    async def estimate(
        self, shots: list[AttributesInfo[BodyEstimatedAttributes]], estimator: AggregatedBodyDescriptorEstimationBroker
    ) -> BodyDescriptor | None:
        if not shots:
            return None

        warps = [shot.attributes.warp.warpedImage for shot in shots]
        _, descriptor = await estimator.estimate(warps)
        return descriptor

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Run body descriptor estimation on the finalized events."""

        estimator = ctx.sdk.engine.estimators.getAggreagtedBodyDescriptorEstimator()

        events, shots = getEventsAndShots(humanTracks, "body")
        results = await asyncio.gather(*map(partial(self.estimate, estimator=estimator), shots))
        for event, descriptor in zip(events, results):
            event.aggregated.aggregatedEstimations.body.descriptor = descriptor


class NodeBodyAttributes(NodeAggregation):
    """Body attributes estimation node."""

    name = "aggregated_body_attributes"
    requires = [NodeBodyWarp, NodeHumanTrackingEventPolicy]

    async def estimate(
        self, shots: list[AttributesInfo[BodyEstimatedAttributes]], estimator: BodyAttributesEstimationBroker
    ) -> BodyAttributes | None:
        if not shots:
            return None

        warps = [shot.attributes.warp.warpedImage for shot in shots]
        res = await asyncio.gather(*map(estimator.estimate, warps))
        return estimator._workers[0]._estimator.aggregate(res)

    async def _run(self, humanTracks: list[HumanTrack], ctx: AnalyticsCtx):
        """Run body attributes estimation on the finalized events."""

        estimator = ctx.sdk.engine.estimators.getBodyAttributesEstimator()

        events, shots = getEventsAndShots(humanTracks, "body")
        results = await asyncio.gather(*map(partial(self.estimate, estimator=estimator), shots))
        for event, attributes in zip(events, results):
            event.aggregated.aggregatedEstimations.body.attributes = attributes


class NodeOverview(NodeAggregation):
    """Overview node."""

    name = "overview"
    requires = [NodeHumanTrackingEventPolicy]

    def __init__(self):
        super().__init__()
        self.frameSyncer = Syncer()

    def scaleDetection(self, detection: FaceDetection | BodyDetection, ratio: float) -> Rect:
        """Scale detection rect according to snapshot ratio."""

        return Rect(
            x=round(detection.boundingBox.rect.x * ratio),
            y=round(detection.boundingBox.rect.y * ratio),
            width=round(detection.boundingBox.rect.width * ratio),
            height=round(detection.boundingBox.rect.height * ratio),
        )

    @nodeMonitoring()
    async def run(self, frameCtx: FrameCtx, ctx: AnalyticsCtx):
        """Run snapshot preserving."""

        if frameCtx.processed is None or frameCtx.processed.humanTracks is None:
            return
        humanTracks = frameCtx.processed.humanTracks

        async with self.frameSyncer.process(frameCtx.number) as lock:
            async with lock.acquire():
                for track in humanTracks:
                    if not (event := track.currentEvent):
                        continue
                    if event.eventStatus is EventStatus.finished:
                        attributes = event.eventShots
                        if attributes.faceShots and attributes.bodyShots:
                            for faceShot in attributes.faceShots:
                                for bodyShot in attributes.bodyShots:
                                    if faceShot.frameOffset == bodyShot.frameOffset:
                                        break
                                if faceShot.frameOffset == bodyShot.frameOffset:
                                    break
                            else:
                                faceShot, bodyShot = attributes.faceShots[0], None
                        elif attributes.faceShots:
                            faceShot, bodyShot = attributes.faceShots[0], None
                        elif attributes.bodyShots:
                            faceShot, bodyShot = None, attributes.bodyShots[0]
                        else:
                            return
                        baseShot = faceShot or bodyShot
                        snapshot, ratio = await toBytes(
                            baseShot.attributes.detection.image, ctx.params.imageRetainPolicy
                        )
                        overview = ShotInfo(baseShot.timeOffset, snapshot)
                        if faceShot:
                            score = faceShot.attributes.detection.boundingBox.score
                            overview.faceDetection = (self.scaleDetection(faceShot.attributes.detection, ratio), score)
                        if bodyShot:
                            score = bodyShot.attributes.detection.boundingBox.score
                            overview.bodyDetection = (self.scaleDetection(bodyShot.attributes.detection, ratio), score)
                        event.aggregated.overview = overview


class _NodeAggregationPreserver:
    """Aggregations preserver node."""

    name = "aggregations_preserver"
    requires = [NodeHumanTrackingEventPolicy]
    follows = [
        NodeOverview,
        NodeLivenessV1,
        NodeDeepfake,
        NodeFaceShots,
        NodeFaceBestshot,
        NodeFaceBasicAttributes,
        NodeFaceDescriptor,
        NodeBodyShots,
        NodeBodyBestshot,
        NodeBodyDescriptor,
        NodeBodyAttributes,
    ]

    async def run(self, frameCtx: FrameCtx, ctx: AnalyticsCtx):
        """Preserve aggregated estimation results."""
        if frameCtx.processed is None or frameCtx.processed.humanTracks is None:
            return
        for track in frameCtx.processed.humanTracks:
            if not (event := track.currentEvent):
                continue
            if event.aggregated:
                if ctx.accumulating:
                    for frame in reversed(event.frames):
                        if frame.frameOffset == frameCtx.number:
                            frame.aggregated = event.aggregated.aggregatedEstimations.asDict()
                            if event.aggregated.overview:
                                frame.overview = event.aggregated.overview.asDict()
                            break
