"""Human tracking analytics."""

import attr
import numpy as np
from lunavl.sdk.image_utils.image import VLImage
from sdk_loop.estimators.te.human_tracker import TEBroker

from luna_analytics_manager import FrameEnviron
from luna_analytics_manager.templates.models import FrequencyParams
from luna_analytics_manager.templates.sdk import SDKAnalyticsCtx

from .frame_structures import FrameCtx, FrameRes, HumanTrackingEvent
from .models import BODY_TARGETS, FACE_TARGETS, Params, Targets, detectorTypeVsTargetsValidation

FrameId = int
TrackId = str


async def initialize(**kwargs):
    pass


class AnalyticsResult:
    """Analytic result on all tracks."""

    def __init__(self):
        self.finalizedTracks: set[TrackId] = set()
        self.finalizedEvents: list[HumanTrackingEvent] = list()

    def finalizeEvent(self, event: HumanTrackingEvent, accumulating: bool):
        """
        Captures event to finalizedEvents map in case if `accumulating` is set to True
        Args:
            event (HumanTrackingEvent): Human tracking event to finalize
            accumulating (bool): Whether to accumulate the estimation result.
        """
        if accumulating:
            self.finalizedEvents.append(event)

    def asDict(self):
        """Build results in API format."""

        result = []
        for event in self.finalizedEvents:
            frames = sorted(event.frames, key=lambda x: x.frameOffset)
            event = {
                "event_id": event.eventId,
                "track_id": event.trackId,
                "video_segment": event.aggregated.videoSegment.asDict(),
                "frames_estimations": [frame.asDict(minimal=False) for frame in frames],
                "aggregated_estimations": frames[-1].aggregated,
            }
            if frames[-1].overview:
                event["overview"] = frames[-1].overview
            result.append(event)
        return {"tracking": {"events": result, "track_count": len(self.finalizedTracks)}}


@attr.dataclass(slots=True)
class AnalyticsCtx(SDKAnalyticsCtx):
    """Analytics context."""

    name: str = "human_tracking"
    params: Params = attr.field(factory=Params)
    targets: Targets = attr.field(factory=Targets)
    aggregated: AnalyticsResult = attr.field(factory=AnalyticsResult)

    client: TEBroker = attr.field(init=False)
    streamId: int = attr.field(init=False)

    _faceTargets: list[str] = attr.field(init=False, default=FACE_TARGETS)
    _bodyTargets: list[str] = attr.field(init=False, default=BODY_TARGETS)

    def __attrs_post_init__(self):
        detectorTypeVsTargetsValidation(
            targets=self.targets,
            params=self.params,
            allowedFaceTargets=self._faceTargets,
            allowedBodyTargets=self._bodyTargets,
        )
        super().__attrs_post_init__()

    async def prepareFrameCtx(self, frameEnv: FrameEnviron[np.ndarray, VLImage]) -> FrameCtx:
        """Prepare frame processing context."""

        return FrameCtx(
            image=frameEnv.frame,
            number=frameEnv.raw.number,
            timestamp=frameEnv.raw.timestamp,
            lastFrame=frameEnv.raw.lastFrame,
            processed=FrameRes(frameEnv.raw.timestamp),
        )

    def getAnalyticsRate(self) -> FrequencyParams:
        """Get frame rate config to be applied by manager."""

        return self.params.rate
