"""Human tracking analytics extra estimations."""

# mypy: disable-error-code="union-attr"

import asyncio

from lunavl.sdk.errors.exceptions import LunaSDKException
from lunavl.sdk.trackengine.structures import BodyTrack, FaceTrack
from sdk_loop.estimators.body_estimators.body_warp_estimator import BodyWarpEstimationBroker
from sdk_loop.estimators.face_estimators.ags_estimator import AGSEstimationBroker
from sdk_loop.estimators.face_estimators.emotions_estimator import EmotionsEstimationBroker
from sdk_loop.estimators.face_estimators.eyes_estimator import EyesEstimationBroker
from sdk_loop.estimators.face_estimators.face_warp_estimator import FaceWarpEstimationBroker
from sdk_loop.estimators.face_estimators.face_warp_quality_estimator import FaceWarpQualityEstimationBroker
from sdk_loop.estimators.face_estimators.gaze_direction_estimator import GazeEstimationBroker
from sdk_loop.estimators.face_estimators.glasses_estimator import GlassesEstimationBroker
from sdk_loop.estimators.face_estimators.head_pose_estimator import HeadPoseEstimationBroker
from sdk_loop.estimators.face_estimators.landmarks_estimator import Landmarks5EstimationBroker
from sdk_loop.estimators.face_estimators.mask_estimator import MaskEstimationBroker
from sdk_loop.estimators.face_estimators.mouth_estimator import MouthStateEstimationBroker
from sdk_loop.estimators.worker import EstimationBrokerBase

from luna_analytics_manager.monitoring.node_monitoring import nodeMonitoring

from .analytics import AnalyticsCtx
from .frame_structures import FrameCtx, HumanTrack
from .human_tracking import NodeHumanTracking


class NodeExtraEstimation:
    """Extra estimation node."""

    name: str
    estimator: EstimationBrokerBase

    def createEstimator(self, engine):
        """Create estimator."""

    async def estimate(self, track: HumanTrack):
        """Perform estimation for track detection."""

    def filterTracks(self, track: HumanTrack) -> bool:
        """Filter tracks for estimation."""

        if not isinstance(track.face, FaceTrack):
            return False
        if track.face.detection is None:
            return False
        return True

    @nodeMonitoring()
    async def run(self, frameCtx: FrameCtx, ctx: AnalyticsCtx):
        """Run estimation on frame tracks."""

        if frameCtx.processed is None or frameCtx.processed.humanTracks is None:
            return

        self.createEstimator(ctx.sdk.engine)
        try:
            await asyncio.gather(*map(self.estimate, filter(self.filterTracks, frameCtx.processed.humanTracks)))
        except LunaSDKException as err:
            frameCtx.error = err.error


class NodeLandmarks5(NodeExtraEstimation):
    """Landmark estimation node."""

    name = "landmarks5"
    requires = [NodeHumanTracking]

    estimator: Landmarks5EstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getLandmarks5Estimator()

    async def estimate(self, track: HumanTrack):
        detection = track.estimated.face.detection
        if not detection.landmarks5:
            detection.landmarks5 = await self.estimator.estimate(detection)
        track.estimated.face.landmarks5 = detection.landmarks5


class NodeHeadpose(NodeExtraEstimation):
    """Headpose estimation node."""

    name = "head_pose"
    requires = [NodeHumanTracking]

    estimator: HeadPoseEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getHeadPoseEstimator()

    async def estimate(self, track: HumanTrack):
        detection = track.estimated.face.detection
        track.estimated.face.headpose = await self.estimator.estimate(detection)


class NodeAGS(NodeExtraEstimation):
    """Approximate garbage score estimation node."""

    name = "ags"
    requires = [NodeHumanTracking]

    estimator: AGSEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getAGSEstimator()

    async def estimate(self, track: HumanTrack):
        detection = track.estimated.face.detection
        track.estimated.face.ags = await self.estimator.estimate(detection)


class NodeFaceWarp(NodeExtraEstimation):
    """Face warp estimation node."""

    name = "face_warp"
    requires = [NodeLandmarks5, NodeHumanTracking]

    estimator: FaceWarpEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getFaceWarpEstimator()

    async def estimate(self, track: HumanTrack):
        detection = track.estimated.face.detection
        track.estimated.face.warp = await self.estimator.estimate(detection)


class NodeFaceWarpQuality(NodeExtraEstimation):
    """Warp quality estimation node."""

    name = "warp_quality"
    requires = [NodeFaceWarp]

    estimator: FaceWarpQualityEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getFaceWarpQualityEstimator()

    async def estimate(self, track: HumanTrack):
        warp = track.estimated.face.warp.warp
        track.estimated.face.warpQuality = await self.estimator.estimate(warp)


class NodeEmotions(NodeExtraEstimation):
    """Emotions estimation node."""

    name = "emotions"
    requires = [NodeFaceWarp]

    estimator: EmotionsEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getEmotionsEstimator()

    async def estimate(self, track: HumanTrack):
        warp = track.estimated.face.warp.warp
        track.estimated.face.emotions = await self.estimator.estimate(warp)


class NodeMask(NodeExtraEstimation):
    """Mask estimation node."""

    name = "mask"
    requires = [NodeFaceWarp]

    estimator: MaskEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getMaskEstimator()

    async def estimate(self, track: HumanTrack):
        warp = track.estimated.face.warp.warp
        track.estimated.face.mask = await self.estimator.estimate(warp)


class NodeGlasses(NodeExtraEstimation):
    """Glasses estimation node."""

    name = "glasses"
    requires = [NodeFaceWarp]

    estimator: GlassesEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getGlassesEstimator()

    async def estimate(self, track: HumanTrack):
        warp = track.estimated.face.warp.warp
        track.estimated.face.glasses = await self.estimator.estimate(warp)


class NodeEyes(NodeExtraEstimation):
    """Eyes estimation node."""

    name = "eyes"
    requires = [NodeFaceWarp]

    estimator: EyesEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getEyesEstimator()

    async def estimate(self, track: HumanTrack):
        warp = track.estimated.face.warp
        track.estimated.face.eyes = await self.estimator.estimate(warp)


class NodeGaze(NodeExtraEstimation):
    """Gaze estimation node."""

    name = "gaze"
    requires = [NodeFaceWarp]

    estimator: GazeEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getGazeDirectionEstimator()

    async def estimate(self, track: HumanTrack):
        warp = track.estimated.face.warp
        track.estimated.face.gaze = await self.estimator.estimate(warp)


class NodeMouth(NodeExtraEstimation):
    """Mouth estimation node."""

    name = "mouth"
    requires = [NodeFaceWarp]

    estimator: MouthStateEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getMouthStateEstimator()

    async def estimate(self, track: HumanTrack):
        warp = track.estimated.face.warp.warp
        track.estimated.face.mouth = await self.estimator.estimate(warp)


class NodeBodyWarp(NodeExtraEstimation):
    """Body warp estimation node."""

    name = "body_warp"
    requires = [NodeHumanTracking]

    estimator: BodyWarpEstimationBroker

    def createEstimator(self, engine):
        self.estimator = engine.estimators.getBodyWarpEstimator()

    def filterTracks(self, track: HumanTrack) -> bool:
        """Filter tracks for estimation."""

        if not isinstance(track.body, BodyTrack):
            return False
        if track.body.detection is None:
            return False
        return True

    async def estimate(self, track: HumanTrack):
        detection = track.estimated.body.detection
        track.estimated.body.warp = await self.estimator.estimate(detection)


class _NodeEstimationPreserver:
    """Extra estimation preserver node."""

    name = "estimations_preserver"
    requires = [NodeHumanTracking]
    follows = [
        NodeLandmarks5,
        NodeHeadpose,
        NodeAGS,
        NodeFaceWarp,
        NodeFaceWarpQuality,
        NodeEmotions,
        NodeMask,
        NodeGlasses,
        NodeEyes,
        NodeGaze,
        NodeMouth,
        NodeMouth,
        NodeBodyWarp,
    ]

    async def run(self, frameCtx: FrameCtx, ctx: AnalyticsCtx):
        """Preserve frame estimation results."""

        if not ctx.accumulating:
            return

        if frameCtx.processed is None or frameCtx.processed.humanTracks is None:
            return

        for track in frameCtx.processed.humanTracks:
            if track.estimated:
                track.trackFrames[-1].estimated = track.estimated.asDict()
