import operator
from uuid import uuid4

import attr
from shapely import Point

from luna_analytics_manager.monitoring.node_monitoring import nodeMonitoring
from luna_analytics_manager.templates.human_tracking import Params
from luna_analytics_manager.templates.human_tracking.analytics import AnalyticsCtx, TrackId
from luna_analytics_manager.templates.human_tracking.estimations import _NodeEstimationPreserver
from luna_analytics_manager.templates.human_tracking.frame_structures import (
    AttributesInfo,
    BodyEstimatedAttributes,
    EventAggregations,
    FaceEstimatedAttributes,
    FrameCtx,
    FrameInfo,
    HumanTrack,
    HumanTrackingEvent,
)
from luna_analytics_manager.templates.human_tracking.human_tracking import NodeHumanTracking
from luna_analytics_manager.templates.image_helpers.roi import scaleDROI
from luna_analytics_manager.templates.models import EventStatus, TrackStatus

EventId = str


@attr.dataclass(slots=True)
class EventInfo:
    """Internal structure to keep event cache data from one frame to another."""

    currentEventId: EventId
    currentEventStatus: EventStatus
    lastEventTime: float = 0
    eventFrames: list[FrameInfo] = attr.field(factory=list)
    eventShots: EventAggregations = attr.field(factory=EventAggregations)


class NodeHumanTrackingEventPolicy:

    name = "event_policy"
    requires = [NodeHumanTracking]
    follows = [_NodeEstimationPreserver]

    def __init__(self):
        self.eventsCache: dict[TrackId, EventInfo] = {}

    @staticmethod
    def checkFaceDetection(detection: FaceEstimatedAttributes, config: Params):
        """Makes sure the face detection satisfies the filters set in analytics parameters."""

        if not detection.detection:
            return False
        filters = config.faceSamples.filters
        if (ags := detection.ags) is not None:
            if (score := filters.score) is not None:
                if ags < score:
                    return False

        if (droi := config.faceSamples.filters.droi) is not None:
            center = detection.detection.boundingBox.rect.center
            if not scaleDROI(detection.detection.image.rect, droi).covers(Point(center.x, center.y)):
                return False

        if (headpose := detection.headpose) is not None:
            if (headRoll := filters.headRoll) is not None:
                if abs(headpose.roll) > headRoll:
                    return False
            if (headPitch := filters.headPitch) is not None:
                if abs(headpose.pitch) > headPitch:
                    return False
            if (headYaw := filters.headYaw) is not None:
                if abs(headpose.yaw) > headYaw:
                    return False
        return True

    @staticmethod
    def checkBodyDetection(detection: BodyEstimatedAttributes, config: Params):
        """Makes sure the body detection satisfies the filters set in analytics parameters."""

        if not detection.detection:
            return False

        if config.bodySamples.filters.score is not None:
            if detection.detection.boundingBox.score < config.bodySamples.filters.score:
                return False

        if (droi := config.bodySamples.filters.droi) is not None:
            center = detection.detection.boundingBox.rect.center
            if not scaleDROI(detection.detection.image.rect, droi).covers(Point(center.x, center.y)):
                return False

        return True

    def collectFaceShots(self, frameCtx: FrameCtx, event: HumanTrackingEvent, config: Params):
        """Makes sure the face detection is good enough to be included in event shots for further aggregation."""

        if not event.face:
            return

        if self.checkFaceDetection(event.estimated.face, config):
            shot = AttributesInfo(frameCtx.number, frameCtx.timestamp, event.estimated.face)
            eventShots = self.eventsCache[event.trackId].eventShots
            if len(eventShots.faceShots) < config.faceSamples.count:
                eventShots.faceShots.append(shot)
            else:
                shots = [*eventShots.faceShots, shot]
                shots.sort(
                    key=operator.attrgetter("attributes.ags", "attributes.detection.boundingBox.score"), reverse=True
                )
                eventShots.faceShots = shots[: config.faceSamples.count]

    def collectBodyShots(self, frameCtx: FrameCtx, event: HumanTrackingEvent, config: Params):
        """Makes sure the body detection is good enough to be included in event shots for further aggregation."""

        if not event.body:
            return

        if self.checkBodyDetection(event.estimated.body, config):
            shot = AttributesInfo(frameCtx.number, frameCtx.timestamp, event.estimated.body)
            eventShots = self.eventsCache[event.trackId].eventShots
            if len(eventShots.bodyShots) < config.bodySamples.count:
                eventShots.bodyShots.append(shot)
            else:
                shots = [*eventShots.bodyShots, shot]
                shots.sort(key=operator.attrgetter("attributes.detection.boundingBox.score"), reverse=True)
                eventShots.bodyShots = shots[: config.bodySamples.count]

    def applyEventPolicy(
        self, track: HumanTrack, timestamp: float, ctx: AnalyticsCtx, isLastFrame: bool
    ) -> EventStatus:
        """
        Checks whether it's time to finish the event based on track status,
        event policy trigger and shots count settings.

        Args:
            track (HumanTrack): Track data.
            timestamp (float): Timestamp of the current frame.
            ctx (AnalyticsCtx): Analytics context.
            isLastFrame (bool): Whether the last frame is being processed.
        Returns:
            EventStatus after applying event policy rules.
        """

        eventPolicy = ctx.params.eventPolicy
        event = track.currentEvent
        eventCache = self.eventsCache[track.trackId]
        attributes = eventCache.eventShots

        if event.trackStatus == TrackStatus.finished or isLastFrame:  # type: ignore
            if attributes.faceShots or attributes.bodyShots:
                event.eventStatus = EventStatus.finished  # type: ignore
                event.trackStatus = TrackStatus.finished  # type: ignore

        elif eventPolicy.trigger == "start":
            # check whether we have collected enough shots.
            # if not, keeps collecting until either it be enough, or track finished
            if not ctx.targets.face or ctx.params.faceSamples.count == len(attributes.faceShots):
                if not ctx.targets.body or ctx.params.bodySamples.count == len(attributes.bodyShots):
                    event.eventStatus = EventStatus.finished  # type: ignore

        elif eventPolicy.trigger == "period":
            offset = eventCache.lastEventTime or 0
            if timestamp - offset >= eventPolicy.interval:
                # checks that there are any detections to don't proceed with an empty event
                if attributes.faceShots or attributes.bodyShots:
                    event.eventStatus = EventStatus.finished  # type: ignore

        return event.eventStatus  # type: ignore

    def processEvent(self, track: HumanTrack, frameCtx: FrameCtx, ctx: AnalyticsCtx) -> None:
        """
        Manages event according to the EventPolicy. In case if event status becomes 'finished',
        it finalizes the event, captures its data to analytics result and either saves its data to local cache
        so the further events processing can rely on it, or clears out cache data in case if track is finished.

        Args:
            track (HumanTrack): Track containing the event to process.
            frameCtx (FrameCtx): FrameCtx containing context of the current frame.
            ctx (AnalyticsCtx): Analytics context.
        """
        eventCache = self.eventsCache[track.trackId]
        event = track.currentEvent
        trackEnded = track.trackStatus == TrackStatus.finished or frameCtx.lastFrame
        eventStatus = self.applyEventPolicy(track, frameCtx.timestamp, ctx, frameCtx.lastFrame)

        if eventStatus == EventStatus.finished:
            event.eventShots = eventCache.eventShots  # type: ignore
            event.finalize()  # type: ignore
            ctx.aggregated.finalizeEvent(event, ctx.accumulating)  # type: ignore
            if not trackEnded:
                eventCache.lastEventTime = frameCtx.timestamp
                eventCache.currentEventStatus = eventStatus

        if trackEnded:
            self.eventsCache.pop(track.trackId)

    def registerEvent(self, track: HumanTrack, ctx: AnalyticsCtx) -> HumanTrackingEvent | None:
        """
        Restores previous event data from cache if any, either creates cache item. Registers event within the track.
        Collects frames the event appears on to perform aggregation when needed.

        Returns:
            HumanTrackingEvent - when event is successfully registered.
            None - in case if previously lasted event has to be the last one,
                   and no more events suppose to be created.
        """
        if not (cache := self.eventsCache.get(track.trackId)):
            event = EventInfo(
                currentEventId=str(uuid4()),
                currentEventStatus=EventStatus.started,
            )
            cache = self.eventsCache[track.trackId] = event
        elif cache.currentEventStatus is EventStatus.finished:
            match ctx.params.eventPolicy.trigger:
                case "start":
                    return None  # no events after the one finished on trigger `start`
                case "period":
                    lastEventTime = cache.lastEventTime
                    event = EventInfo(
                        currentEventId=str(uuid4()), currentEventStatus=EventStatus.started, lastEventTime=lastEventTime
                    )
                    cache = self.eventsCache[track.trackId] = event
        else:
            cache.currentEventStatus = EventStatus.inProcess
        cache.eventFrames.extend(track.trackFrames)
        track.registerEvent(cache.currentEventId, cache.currentEventStatus, cache.eventFrames)
        return track.currentEvent

    @nodeMonitoring()
    async def run(self, frameCtx: FrameCtx, ctx: AnalyticsCtx):
        """
        Manages to apply event policy rules to processed frames
        and to finalize and aggregate the finished events if any.
        """
        if frameCtx.processed is None or frameCtx.processed.humanTracks is None:
            return
        for track in frameCtx.processed.humanTracks:
            if not (event := self.registerEvent(track, ctx)):
                continue
            self.collectBodyShots(frameCtx, event, ctx.params)
            self.collectFaceShots(frameCtx, event, ctx.params)
            self.processEvent(track, frameCtx, ctx)
