"""Human tracking frame models."""

from typing import Generic, TypeVar

import attr
from lunavl.sdk.descriptors.descriptors import BodyDescriptor, FaceDescriptor
from lunavl.sdk.detectors.bodydetector import BodyDetection
from lunavl.sdk.detectors.facedetector import FaceDetection, Landmarks5
from lunavl.sdk.estimators.body_estimators.body_attributes import BodyAttributes
from lunavl.sdk.estimators.body_estimators.bodywarper import BodyWarp
from lunavl.sdk.estimators.face_estimators.basic_attributes import BasicAttributes
from lunavl.sdk.estimators.face_estimators.deepfake import Deepfake
from lunavl.sdk.estimators.face_estimators.emotions import Emotions
from lunavl.sdk.estimators.face_estimators.eyes import EyesEstimation, GazeDirection, WarpWithLandmarks5
from lunavl.sdk.estimators.face_estimators.glasses import Glasses
from lunavl.sdk.estimators.face_estimators.head_pose import HeadPose
from lunavl.sdk.estimators.face_estimators.livenessv1 import LivenessV1
from lunavl.sdk.estimators.face_estimators.mask import Mask
from lunavl.sdk.estimators.face_estimators.mouth_state import MouthStates
from lunavl.sdk.estimators.face_estimators.warp_quality import Quality
from lunavl.sdk.image_utils.geometry import Rect
from lunavl.sdk.trackengine.structures import BodyTrack, FaceTrack

from luna_analytics_manager.templates.models import EventStatus, TrackStatus, VideoSegment
from luna_analytics_manager.templates.sdk import SDKFrameCtx

from .adaptors import MinimalTrack, buildBodyAttributes, buildTrack


@attr.dataclass(slots=True)
class FrameInfo:
    """Frame info for single track."""

    frameOffset: int
    timeOffset: float
    face: MinimalTrack | None = None
    body: MinimalTrack | None = None
    estimated: dict = attr.field(factory=dict)
    aggregated: dict = attr.field(factory=dict)
    overview: dict = attr.field(factory=dict)

    def addTrack(self, face: FaceTrack | MinimalTrack, body: BodyTrack | MinimalTrack) -> "FrameInfo":
        """Add human tracking info."""

        self.face = (face if isinstance(face, MinimalTrack) else MinimalTrack(face)) if face else None
        self.body = (body if isinstance(body, MinimalTrack) else MinimalTrack(body)) if body else None
        return self

    def asDict(self, minimal: bool = True):
        """Build results in API format."""

        res = {
            "time_offset": self.timeOffset,
            "estimated": {
                "face": (
                    {"detection": buildTrack(self.face)} | (self.estimated["face"] if not minimal else {})
                    if self.face
                    else None
                ),
                "body": (
                    {"detection": buildTrack(self.body)} | (self.estimated["body"] if not minimal else {})
                    if self.body
                    else None
                ),
            },
        }
        return res


@attr.dataclass(slots=True)
class FaceEstimatedAttributes:
    """Estimated attributes."""

    detection: FaceDetection | None = None
    landmarks5: Landmarks5 | None = None
    headpose: HeadPose | None = None
    ags: float | None = None
    warp: WarpWithLandmarks5 | None = None
    warpQuality: Quality | None = None
    emotions: Emotions | None = None
    mask: Mask | None = None
    glasses: Glasses | None = None
    eyes: EyesEstimation | None = None
    gaze: GazeDirection | None = None
    mouth: MouthStates | None = None

    def __bool__(self):
        return self.detection is not None

    def asDict(self):
        """Build results in API format."""

        res = {}
        if self.landmarks5 is not None:
            res["landmarks5"] = self.landmarks5.asDict()
        if self.headpose is not None:
            res["head_pose"] = self.headpose.asDict()
        if self.ags is not None:
            res["ags"] = self.ags
        if self.warpQuality is not None:
            res["warp_quality"] = self.warpQuality.asDict()
        if self.emotions is not None:
            res["emotions"] = self.emotions.asDict()
        if self.mask is not None:
            res["mask"] = self.mask.asDict()
        if self.glasses is not None:
            res["glasses"] = self.glasses.asDict()
        if self.eyes is not None:
            res["eyes"] = {
                "left_eye": {
                    "state": self.eyes.leftEye.state.name.lower(),
                    "iris_landmarks": self.eyes.leftEye.irisLandmarks.asDict(),
                },
                "right_eye": {
                    "state": self.eyes.rightEye.state.name.lower(),
                    "iris_landmarks": self.eyes.rightEye.irisLandmarks.asDict(),
                },
            }
        if self.gaze is not None:
            res["gaze"] = self.gaze.asDict()
        if self.mouth is not None:
            mouth = self.mouth.asDict()
            mouth["score"] = 1.0
            del mouth["properties"]
            res["mouth"] = mouth
        return res


@attr.dataclass(slots=True)
class BodyEstimatedAttributes:
    """Estimated attributes."""

    detection: BodyDetection | None = None
    warp: BodyWarp | None = None

    def __bool__(self):
        return self.detection is not None

    def asDict(self):
        """Build results in API format."""

        return {}


@attr.dataclass(slots=True)
class ShotInfo:
    """Shot extended with info about frame."""

    timeOffset: float
    snapshot: bytes
    faceDetection: tuple[Rect, float] | None = None
    bodyDetection: tuple[Rect, float] | None = None

    def asDict(self):
        res = {
            "time_offset": self.timeOffset,
            "image": self.snapshot,
        }
        if self.faceDetection or self.bodyDetection:
            res["detection"] = {
                "face": (
                    {"bbox": self.faceDetection[0].asDict(), "score": self.faceDetection[1]}
                    if self.faceDetection
                    else None
                ),
                "body": (
                    {"bbox": self.bodyDetection[0].asDict(), "score": self.bodyDetection[1]}
                    if self.bodyDetection
                    else None
                ),
            }
        return res


@attr.dataclass(slots=True)
class EstimatedAttributes:
    """Estimated attributes."""

    face: FaceEstimatedAttributes = attr.field(factory=FaceEstimatedAttributes)
    body: BodyEstimatedAttributes = attr.field(factory=BodyEstimatedAttributes)

    def asDict(self):
        """Build results in API format."""

        return {"face": self.face.asDict(), "body": self.body.asDict()}


ATTRIBUTES = TypeVar("ATTRIBUTES")


@attr.dataclass(slots=True)
class AttributesInfo(Generic[ATTRIBUTES]):
    """Attributes extended with frame info estimations."""

    frameOffset: int
    timeOffset: float
    attributes: ATTRIBUTES


@attr.dataclass(slots=True)
class EventAggregations:
    """Event attributes to be used for aggregation."""

    faceShots: list[AttributesInfo[FaceEstimatedAttributes]] = attr.field(factory=list)
    bodyShots: list[AttributesInfo[BodyEstimatedAttributes]] = attr.field(factory=list)


@attr.dataclass(slots=True)
class HumanTrackingEvent:
    """Event info for a single frame."""

    @attr.dataclass(slots=True)
    class AggregatedEvent:
        """Aggregated event data."""

        @attr.dataclass(slots=True)
        class Attributes:
            """Attributes for event aggregation."""

            @attr.dataclass(slots=True)
            class FaceAttributes:
                """Face aggregated attributes."""

                shots: list[ShotInfo] | None = None
                bestshot: ShotInfo | None = None
                descriptor: FaceDescriptor | None = None
                basicAttributes: BasicAttributes | None = None
                liveness: LivenessV1 | None = None
                deepfake: Deepfake | None = None

                def asDict(self):
                    """Build results in API format."""

                    res = {}
                    if self.shots is not None:
                        res["samples"] = [shot.asDict() for shot in self.shots]
                    if self.bestshot is not None:
                        res["thumbnail"] = self.bestshot.asDict()
                    if self.descriptor is not None:
                        res["descriptor"] = {
                            "sdk_descriptor": self.descriptor.rawDescriptor,
                            "score": self.descriptor.garbageScore,
                        }
                    if self.basicAttributes is not None:
                        res["basic_attributes"] = self.basicAttributes.asDict()
                    if self.liveness is not None:
                        res["liveness"] = self.liveness.asDict()
                    if self.deepfake is not None:
                        res["deepfake"] = self.deepfake.asDict()
                    return res

            @attr.dataclass(slots=True)
            class BodyAttributes:
                """Body aggregated attributes."""

                shots: list[ShotInfo] | None = None
                bestshot: ShotInfo | None = None
                descriptor: BodyDescriptor | None = None
                attributes: BodyAttributes | None = None

                def asDict(self):
                    """Build results in API format."""

                    res = {}
                    if self.shots is not None:
                        res["samples"] = [shot.asDict() for shot in self.shots]
                    if self.bestshot is not None:
                        res["thumbnail"] = self.bestshot.asDict()
                    if self.descriptor:
                        res["descriptor"] = {
                            "sdk_descriptor": self.descriptor.rawDescriptor,
                            "score": self.descriptor.garbageScore,
                        }
                    if self.attributes:
                        res["attributes"] = buildBodyAttributes(self.attributes)
                    return res

            face: FaceAttributes = attr.field(factory=FaceAttributes)
            body: BodyAttributes = attr.field(factory=BodyAttributes)

            def asDict(self):
                """Build results in API format."""

                result = {"face": self.face.asDict() or None, "body": self.body.asDict() or None}
                return result

        trackId: str
        eventId: str
        framesEstimations: list[FrameInfo]
        videoSegment: VideoSegment
        aggregatedEstimations: Attributes = attr.field(factory=Attributes)
        overview: ShotInfo | None = None

        def asDict(self):
            """Build results in API format."""

            result = {
                "track_id": self.trackId,
                "event_id": self.eventId,
                "video_segment": self.videoSegment.asDict(),
                "frames_estimations": [frame.asDict() for frame in self.framesEstimations],
                "aggregated_estimations": self.aggregatedEstimations.asDict(),
            }
            if self.overview:
                result["overview"] = self.overview.asDict()
            return result

    trackId: str
    eventId: str
    trackStatus: TrackStatus
    eventStatus: EventStatus
    face: FaceTrack | MinimalTrack | None
    body: BodyTrack | MinimalTrack | None
    frames: list[FrameInfo] = attr.field(factory=list)
    estimated: EstimatedAttributes = attr.field(factory=EstimatedAttributes)
    aggregated: AggregatedEvent | None = None
    eventShots: EventAggregations = attr.field(factory=EventAggregations)
    eventType: str = "human_tracking"

    def __attrs_post_init__(self):
        if isinstance(self.face, FaceTrack):
            self.estimated.face.detection = self.face.detection
        if isinstance(self.body, BodyTrack):
            self.estimated.body.detection = self.body.detection

    def finalize(self) -> None:
        """Generates event aggregation data based on collected frames"""
        videoSegment = VideoSegment(startTimeOffset=self.frames[0].timeOffset, endTimeOffset=self.frames[-1].timeOffset)
        self.aggregated = self.AggregatedEvent(self.trackId, self.eventId, self.frames, videoSegment)

    def asDict(self):
        """Build results in API format."""

        res = {
            "track_id": self.trackId,
            "event_id": self.eventId,
            "event_status": self.eventStatus.value,
            "track_status": self.trackStatus.value,
            "estimated": {
                "face": {"detection": buildTrack(self.face)} | self.estimated.face.asDict() if self.face else None,
                "body": {"detection": buildTrack(self.body)} | self.estimated.body.asDict() if self.body else None,
            },
        }
        if self.aggregated:
            res["aggregated"] = self.aggregated.asDict()
        return res


@attr.dataclass(slots=True)
class HumanTrack:
    """Human track data"""

    trackId: str
    trackStatus: TrackStatus = TrackStatus.started
    face: FaceTrack | None = None
    body: BodyTrack | None = None
    currentEvent: HumanTrackingEvent | None = None
    trackFrames: list[FrameInfo] = attr.field(factory=list)
    estimated: EstimatedAttributes = attr.field(factory=EstimatedAttributes)

    def __attrs_post_init__(self):
        if self.face:
            self.estimated.face.detection = self.face.detection
        if self.body:
            self.estimated.body.detection = self.body.detection

    def registerEvent(self, eventId: str, eventStatus: EventStatus, frames: list[FrameInfo]) -> None:
        """Registers new event for the track."""
        self.currentEvent = HumanTrackingEvent(
            self.trackId, eventId, self.trackStatus, eventStatus, self.face, self.body, frames, self.estimated
        )

    def finalize(self) -> None:
        """Sets track status to finished."""
        self.trackStatus = TrackStatus.finished

    def asDict(self):
        """Build results in API format."""
        return {
            "track_id": self.trackId,
            "track_status": self.trackStatus.value,
            "face": {"detection": buildTrack(self.face)} | self.estimated.face.asDict() if self.face else None,
            "body": {"detection": buildTrack(self.body)} | self.estimated.body.asDict() if self.body else None,
            "event": self.currentEvent.asDict() if self.currentEvent else None,
        }


@attr.dataclass(slots=True)
class FrameRes:
    """Analytics result on a single frame"""

    timeOffset: float
    humanTracks: list[HumanTrack] | None = None

    @property
    def events(self):
        if self.humanTracks is None:
            return None
        return [track.currentEvent for track in self.humanTracks]

    def asDict(self):
        """Build results in API format."""
        if self.humanTracks is None:
            return {
                "time_offset": self.timeOffset,
                "human_tracks": None,
            }
        return {
            "time_offset": self.timeOffset,
            "human_tracks": [track.asDict() for track in self.humanTracks],
        }


@attr.dataclass(slots=True)
class FrameCtx(SDKFrameCtx):
    """Frame processing context."""

    processed: FrameRes | None = None
