from collections.abc import Hashable
from typing import Generic, Type, TypeVar
from uuid import uuid4

import attr
from lunavl.sdk.errors.exceptions import LunaSDKException
from lunavl.sdk.trackengine.structures import TrackingResult
from sdk_loop.estimators.te.human_tracker import Frame, TEBroker

from luna_analytics_manager import Node
from luna_analytics_manager.cache import CacheDescriptor
from luna_analytics_manager.monitoring.monitoring import NodeMonitoringPoint
from luna_analytics_manager.monitoring.node_monitoring import nodeMonitoring
from luna_analytics_manager.monitoring.storage import NODES_STORAGE
from luna_analytics_manager.sync import Syncer2 as Syncer
from luna_analytics_manager.templates.human_tracking.analytics import AnalyticsCtx, TrackId
from luna_analytics_manager.templates.human_tracking.frame_structures import FrameCtx, FrameInfo, HumanTrack
from luna_analytics_manager.templates.image_helpers.roi import getRelativeROIRect
from luna_analytics_manager.templates.models import RoiPolicy, TrackStatus
from luna_analytics_manager.templates.sdk import SDKFrameCtx

CoreTrackId = int
HumanTrackT = TypeVar("HumanTrackT", bound=HumanTrack)
FrameCtxT = TypeVar("FrameCtxT", bound=SDKFrameCtx)


def hashValue(value) -> int:
    """Hash value or return 0."""
    if value and isinstance(value, Hashable):
        return hash(value)
    return 0


def hashFunc(frameCtx: FrameCtx, ctx: AnalyticsCtx) -> int:
    """Get hash for frameCtx + ctx."""
    number = hash(frameCtx.number)
    image = roi = detectorType = bodyReid = 0

    if img := getattr(frameCtx, "image", None):
        image = hash(img.coreImage)

    if getattr(ctx, "params", None):
        roi = hashValue(ctx.params.roi)
        if getattr(ctx.params, "tracking", None):
            detectorType = hashValue(ctx.params.tracking.detectorType)
            bodyReid = hashValue(ctx.params.tracking.bodyReid)

    return hash((number, image, roi, detectorType, bodyReid))


@attr.dataclass(slots=True)
class TrackInfo:
    """Internal structure to keep track processing side data from frame to frame"""

    framesCount: int = 1
    probed: bool = False


class HumanTrackingNodeMonitoring(NodeMonitoringPoint):
    """Human tracking node monitoring"""

    __slots__ = ("faceCount", "bodyCount", "peopleCount")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.faceCount = None
        self.bodyCount = None
        self.peopleCount = None

    def getFields(self) -> dict:
        """Get monitoring fields"""
        res = super().getFields()
        res["people_count"] = self.peopleCount
        if self.faceCount is not None:
            res["face_count"] = self.faceCount
        if self.bodyCount is not None:
            res["body_count"] = self.bodyCount
        return res


class NodeHumanTrackingBase(Generic[HumanTrackT, FrameCtxT]):
    """Human tracking node."""

    # node name
    name = "human_tracking"
    # requires nodes
    requires: list[Type[Node]] = []
    # factory for generating new track instance
    trackCls: Type[HumanTrackT] = HumanTrack  # type: ignore
    # cache for storing results of identical runs
    cache: CacheDescriptor = CacheDescriptor(hashFunc=hashFunc)

    def __init__(self):
        super().__init__()
        self.monitoring: dict[int, HumanTrackingNodeMonitoring] = {}
        self._frameSyncer = Syncer()
        self._activeTracks: dict[TrackId, TrackInfo] = {}
        self._fsdkTrackIdToTrackIdMap: dict[CoreTrackId, TrackId] = {}
        self._trackIdToFsdkTrackIdMap: dict[TrackId, CoreTrackId] = {}
        self._streamId = -1
        self._client: None | TEBroker = None

    def getHTRoi(self, ctx: AnalyticsCtx) -> RoiPolicy:
        """Get human tracking roi param"""
        return ctx.params.roi

    def _collectMonitoring(self, frameCtx: FrameCtxT, humanTracks: list[HumanTrackT], detectFace, detectBody):
        """Collect human tracking node result for monitoring"""
        if not NODES_STORAGE.enable:
            return
        if not frameCtx.processed.humanTracks:
            return
        point = self.monitoring[frameCtx.number]
        faceCount = 0
        bodyCount = 0
        point.peopleCount = len(humanTracks)
        if detectFace and detectBody:

            for track in humanTracks:
                if track.face is not None:
                    faceCount += 1

                if track.body is not None:
                    bodyCount += 1

            point.faceCount = faceCount
            point.bodyCount = bodyCount
            return

        if detectBody:
            for track in humanTracks:
                if track.body is not None:
                    bodyCount += 1
                point.bodyCount = bodyCount
                return
        if detectFace:
            for track in humanTracks:
                if track.face is not None:
                    faceCount += 1
                point.faceCount = faceCount
                return

    def _registerTrack(self, trackId: CoreTrackId) -> TrackId:
        """
        Register the track in track maps by newly generated track id.

        Args:
            trackId (CoreTrackId): Numeric track id from estimator.
        Returns:
            TrackId: generated track id string.
        """

        newTrackId = str(uuid4())
        self._fsdkTrackIdToTrackIdMap[trackId] = newTrackId
        self._trackIdToFsdkTrackIdMap[newTrackId] = trackId
        self._activeTracks[newTrackId] = TrackInfo()
        return newTrackId

    def _finalizeTrack(self, trackId: TrackId) -> None:
        """
        Clears track data from track maps.

        Args:
            trackId (TrackId): Registered track's id.
        """
        coreTrackId = self._trackIdToFsdkTrackIdMap.pop(trackId)
        self._fsdkTrackIdToTrackIdMap.pop(coreTrackId)
        self._activeTracks.pop(trackId)

    def _finalizeTracks(self, tracks: list[HumanTrackT]) -> None:
        """Clear track data for finished tracks"""
        for track in tracks:
            if track.trackStatus == TrackStatus.finished:
                self._finalizeTrack(track.trackId)

    def getProbeCount(self, ctx: AnalyticsCtx):
        """
        Get probe count from analytics context for HumanTracking node
        """
        return ctx.params.probeCount

    def processLostTracks(
        self, ctx: AnalyticsCtx, frameCtx: FrameCtxT, lostTracks: set[TrackId], endTracks: set[TrackId]
    ) -> list[HumanTrackT]:
        """
        Creates dummy track data for tracks that been lost for some reason but not ended yet
        Fills the lost tracks data with face and body detections from the last successfully estimated frame.

        Args:
            ctx (AnalyticsCtx): Analytics context object.
            frameCtx (FrameCtx): Frame context object.
            lostTracks (set[TrackId]): List of lost track ids.
            endTracks (set[TrackId]): List of end track ids.
        Returns:
            list[HumanTrackT]: List of dummy track data.
        """

        frameTracks = []
        for trackId in lostTracks:
            trackInfo = self._activeTracks[trackId]
            trackInfo.framesCount += 1
            trackEnded = frameCtx.lastFrame or trackId in endTracks
            trackStatus = TrackStatus.finished if trackEnded else TrackStatus.inProcess

            if not self.probeTrack(trackId, self.getProbeCount(ctx)):
                if trackEnded:
                    self._finalizeTrack(trackId)
                continue
            track = self.trackCls(trackId, trackStatus, None, None)  # type: ignore
            frameTracks.append(track)

        return frameTracks

    def _handleEstimation(self, ctx: AnalyticsCtx, frameCtx: FrameCtxT, result: TrackingResult) -> list[HumanTrackT]:
        """
        Registers new tracks in track maps and gathers tracks data from the frame estimations results. Filters out
        tracks that are yet too short. Lost tracks are filled with the data from the previous frame.

        Args:
            result (TrackingResult): Result of frame estimations.
            frameCtx (FrameCtx): Frame context.
            ctx (AnalyticsCtx): Analytics context.
        Returns:
            list[HumanTrackT]: Current frame's tracks list prepared for further processing.
        """
        frameTracks: list[HumanTrackT] = []
        if not result:
            return frameTracks
        receivedTracks = set()
        probeCount = self.getProbeCount(ctx)
        for coreTrack in result.humanTracks:

            if not (trackId := self._fsdkTrackIdToTrackIdMap.get(coreTrack.trackId)):
                trackId = self._registerTrack(coreTrack.trackId)
            else:
                self._activeTracks[trackId].framesCount += 1

            receivedTracks.add(trackId)

            faceTrack = coreTrack.face
            bodyTrack = coreTrack.body
            trackEnded = frameCtx.lastFrame or coreTrack.trackId in result.endTracks

            if not (trackStatus := self.probeTrack(trackId, probeCount)):
                if trackEnded:
                    self._finalizeTrack(trackId)
                continue

            trackStatus = TrackStatus.finished if trackEnded else trackStatus

            track = self.trackCls(trackId, trackStatus, faceTrack, bodyTrack)
            frameTracks.append(track)

        lostTracks = set(self._activeTracks) - receivedTracks
        endTracks = set(map(self._fsdkTrackIdToTrackIdMap.get, result.endTracks))
        missingTracks = self.processLostTracks(ctx, frameCtx, lostTracks, endTracks)  # type: ignore
        frameTracks.extend(missingTracks)

        self._finalizeTracks(frameTracks)

        return frameTracks

    def probeTrack(self, trackId: TrackId, probeCount: int) -> TrackStatus | None:
        """
        Check whether the track length exceeds the min number of frames.
        Marks track as `probed` in active tracks map when its length reaches the minimum number of frames.

        Args:
            trackId (TrackId): ID of the track to probe.
            probeCount (int): The number of frames that is enough to consider the track valid.
        Returns:
            - TrackStatus.started in case if track length reaches the minimum number of frames.
            - TrackStatus.inProcess in case if track has been previously marked as `probed` within active tracks map.
            - None if track length isn't enough to consider the track valid.
        """
        trackInfo = self._activeTracks[trackId]
        if trackInfo.probed:
            return TrackStatus.inProcess
        elif trackInfo.framesCount == probeCount + 1:
            trackInfo.probed = True
            return TrackStatus.started
        return None

    def setNodeResult(self, ctx: AnalyticsCtx, frameCtx: FrameCtxT, tracks: list[HumanTrackT]):
        frameCtx.processed.humanTracks = tracks  # type: ignore

    def setFrameEstimations(self, ctx: AnalyticsCtx, frameCtx: FrameCtxT, frameTracks: list[HumanTrackT]) -> None:
        """
        Finalizes tracks in case of the video's last frame comes either the track status is finished.
        Sets processed tracks' data to the frame context after processing.

        Args:
            ctx (AnalyticsCtx): Analytics Context.
            frameCtx (FrameCtx): Current frame context.
            frameTracks (dict[TrackId, HumanTrack]): Current frame prepared tracks data.
        """
        tracks = []
        for track in frameTracks:

            frameInfo = FrameInfo(frameCtx.number, frameCtx.timestamp).addTrack(track.face, track.body)
            track.trackFrames.append(frameInfo)
            tracks.append(track)

        self.setNodeResult(ctx, frameCtx, tracks)

    async def processTracks(
        self, tracks: list[HumanTrackT], frameCtx: FrameCtxT, ctx: AnalyticsCtx
    ) -> list[HumanTrackT]:
        """
        A tracking processing custom logic.

        Args:
            tracks (HumanTrackT): current tracks.
            ctx (AnalyticsCtx): Analytics Context.
            frameCtx (FrameCtx): Current frame context.
        Returns:
            tracks for following analytics processing
        """
        self.setFrameEstimations(ctx, frameCtx, tracks)
        return tracks

    def getTEFrame(self, frameCtx: FrameCtxT):
        """Get input te Frame structure from current frame ctx"""
        return Frame(
            image=frameCtx.image,
            streamId=self._streamId,
            frameNumber=frameCtx.number,
            lastFrame=frameCtx.lastFrame,
        )

    def configureTE(self, ctx: AnalyticsCtx, frameCtx: FrameCtxT):
        """Configure te for current stream an analytics"""
        detectFace = int(ctx.params.tracking.detectorType in ("face", "human"))
        detectBody = int(ctx.params.tracking.detectorType in ("body", "human"))
        useBodyReId = int(ctx.params.tracking.bodyReid)

        self._client = ctx.sdk.engine.estimators.getHumanTracker(
            detectFace=detectFace, detectBody=detectBody, useBodyReId=useBodyReId
        )
        if self._streamId == -1:
            self._streamId = self._client.registerStream()
            if self.getHTRoi(ctx):
                roi = getRelativeROIRect(frameCtx.image.rect, self.getHTRoi(ctx))
                self._client.configureStream(self._streamId, roi=roi)
        return self._client

    @cache.cache(key="estimation")
    async def estimate(self, frameCtx: FrameCtx, ctx: AnalyticsCtx, client: TEBroker) -> TrackingResult:
        frame = self.getTEFrame(frameCtx)
        return await client.estimate(frame)

    @cache.cache(key="client")
    def getClient(self, frameCtx: FrameCtx, ctx: AnalyticsCtx) -> TEBroker:
        return self._client or self.configureTE(ctx, frameCtx)

    @nodeMonitoring(HumanTrackingNodeMonitoring)
    async def run(self, frameCtx: FrameCtxT, ctx: AnalyticsCtx):
        """Run human tracking."""
        client = self.getClient(frameCtx, ctx)

        async with self._frameSyncer.process(frameCtx.number) as lock:

            result = None

            try:
                result = await self.estimate(frameCtx, ctx, client)
            except LunaSDKException as err:
                frameCtx.error = err.error
            finally:
                async with lock.acquire():
                    frameTracks = self._handleEstimation(ctx, frameCtx, result)
                    frameTracks = await self.processTracks(frameTracks, frameCtx, ctx)

                    teSettings = client.settings.te
                    self._collectMonitoring(frameCtx, frameTracks, teSettings.detectFace, teSettings.detectBody)


class NodeHumanTracking(NodeHumanTrackingBase):
    """
    HumanTracking Node. THis node are using in human tracking analytics (videosdk).
    """

    async def processTracks(self, tracks: list[HumanTrackT], frameCtx: FrameCtx, ctx: AnalyticsCtx):
        """Apply vidoesdk target logic"""
        tracks = await super().processTracks(tracks, frameCtx, ctx)
        for track in tracks:
            if track.trackStatus == TrackStatus.finished:
                ctx.aggregated.finalizedTracks.add(track.trackId)

            if not ctx.targets.body:
                track.body = None

            if not ctx.targets.face:
                track.face = None

        return tracks
