"""Human tracking analytics models."""

from typing import Annotated, Literal

from annotated_types import Ge, Le
from pydantic import AfterValidator, Field, Strict, field_validator
from pydantic_core import InitErrorDetails, PydanticCustomError, ValidationError
from vlutils.structures.pydantic import BaseModel

from luna_analytics_manager.templates.mixins.frame_frequency_mixins import RateFrame, RateTime
from luna_analytics_manager.templates.models import DroiPolicy, WrapModelError, WrapUnionError
from luna_analytics_manager.templates.sdk import SDKAnalyticsParams, SDKAnalyticsTargets

COMMON_TARGETS = ["overview"]
FACE_TARGETS = [
    "face_detection",
    "landmarks5",
    "head_pose",
    "ags",
    "warp_quality",
    "emotions",
    "mask",
    "glasses",
    "eyes",
    "gaze",
    "mouth",
    "aggregated_liveness",
    "aggregated_deepfake",
    "aggregated_face_samples",
    "aggregated_face_thumbnail",
    "aggregated_face_basic_attributes",
    "aggregated_face_descriptor",
]
BODY_TARGETS = [
    "body_detection",
    "aggregated_body_samples",
    "aggregated_body_thumbnail",
    "aggregated_body_attributes",
    "aggregated_body_descriptor",
]
ALL_TARGETS = COMMON_TARGETS + FACE_TARGETS + BODY_TARGETS


class Targets(SDKAnalyticsTargets[Literal[*ALL_TARGETS]]):  # type: ignore[valid-type]
    """Human tracking targets."""

    _default = {"face_detection"}

    @property
    def face(self) -> bool:
        if set(FACE_TARGETS) & self.root:
            return True
        if set(COMMON_TARGETS) & self.root and not self.body:
            return True
        return False

    @property
    def body(self) -> bool:
        if set(BODY_TARGETS) & self.root:
            return True
        return False

    @field_validator("root", mode="after")
    def targetValidator(cls, value):
        """Validate target values."""
        if not value:
            raise PydanticCustomError("ValidationError", "At least one target is required.")
        return value


class FaceShotParams(BaseModel):
    """Face shot parameters."""

    class Filters(BaseModel):
        score: Annotated[float | None, Ge(0), Le(1), Strict()] = 0.6
        headRoll: Annotated[int | None, Ge(0), Le(90), Strict()] = 15
        headPitch: Annotated[int | None, Ge(0), Le(90), Strict()] = 15
        headYaw: Annotated[int | None, Ge(0), Le(90), Strict()] = 15
        droi: DroiPolicy

    count: Annotated[int, Ge(0), Strict()] = 5
    filters: Annotated[Filters, WrapModelError] = Field(default_factory=Filters)


class BodyShotParams(BaseModel):
    """Body shot parameters."""

    class Filters(BaseModel):
        score: Annotated[float | None, Ge(0), Le(1), Strict()] = 0.3
        droi: DroiPolicy

    count: Annotated[int, Ge(0), Strict()] = 5
    filters: Annotated[Filters, WrapModelError] = Field(default_factory=Filters)


class TrackingParams(BaseModel):
    """Tracking parameters."""

    detectorType: Literal["face", "body", "human"] = "face"
    bodyReid: Literal[0, 1] = 0


class ImageRetainPolicy(BaseModel):
    """Special Image Retain Policy without ttl, which is unusable for HT"""

    quality: Annotated[float, Ge(0.0), Le(1.0), Strict(), AfterValidator(lambda x: round(x * 100))] = 100
    maxSize: Annotated[int, Ge(0), Strict()] = 640
    mimetype: Literal["JPEG", "PNG"] = "JPEG"


class Params(SDKAnalyticsParams):
    """Human tracking parameters."""

    rate: Annotated[RateFrame | RateTime, WrapUnionError, Field(discriminator="unit", default=RateFrame(period=1))]
    tracking: Annotated[TrackingParams, WrapModelError] = Field(default_factory=TrackingParams)
    faceSamples: Annotated[FaceShotParams, WrapModelError] = Field(default_factory=FaceShotParams)
    bodySamples: Annotated[BodyShotParams, WrapModelError] = Field(default_factory=BodyShotParams)
    imageRetainPolicy: Annotated[ImageRetainPolicy, WrapModelError, Field(default_factory=ImageRetainPolicy)]


def detectorTypeVsTargetsValidation(
    targets: Targets, params: Params, allowedFaceTargets: list[str], allowedBodyTargets: list[str]
):
    """Validate that detector type compatible with targets"""
    if params.tracking.detectorType == "human":
        return
    try:
        if targets.face and params.tracking.detectorType == "body":
            raise ValueError(f"{allowedFaceTargets} targets must go along with face or human detector types")
        if targets.body and params.tracking.detectorType == "face":
            raise ValueError(f"{allowedBodyTargets} targets must go along with body or human detector types")
    except ValueError as err:
        raise ValidationError.from_exception_data(
            "Incompatible detector type and targets",
            [
                InitErrorDetails(
                    type=PydanticCustomError("Incompatible detector type and targets", str(err)),
                    loc=(),
                    input=params.tracking.detectorType,
                )
            ],
        )
