from lunavl.sdk.image_utils.geometry import Rect
from shapely import Geometry, affinity

from luna_analytics_manager.templates.models import DROI, ROI


def getROIRect(sourceRect: Rect, targetRect: ROI) -> Rect:
    """
    Get ROI from source image

    Args:
        sourceRect: source rect
        targetRect: target rect
    Returns:
        ROI of the source image

    Raises:
        ValueError: If mode value is unsupported
    """
    match targetRect.mode:
        case "abs":
            return Rect(targetRect.x, targetRect.y, targetRect.width, targetRect.height)
        case "percent":
            imgHeight, imgWidth = sourceRect.height, sourceRect.width
            x, width = round(targetRect.x * imgWidth / 100), round(targetRect.width * imgWidth / 100)
            y, height = round(targetRect.y * imgHeight / 100), round(targetRect.height * imgHeight / 100)
            return Rect(x, y, width, height)
        case _:
            raise ValueError(f"ROI/DROI '{targetRect.mode}' mode is unsupported")


def getRelativeROIRect(sourceRect: Rect, targetRect: ROI) -> Rect:
    """
    Get ROI from source image

    Args:
        sourceRect: source rect
        targetRect: target rect
    Returns:
        ROI of the source image

    Raises:
        ValueError: If mode value is unsupported
    """
    match targetRect.mode:
        case "abs":
            imgHeight, imgWidth = sourceRect.height, sourceRect.width
            x, width = targetRect.x / imgWidth, targetRect.width / imgWidth
            y, height = targetRect.y / imgHeight, targetRect.height / imgHeight
            return Rect(x, y, width, height)
        case "percent":
            return Rect(targetRect.x / 100, targetRect.y / 100, targetRect.width / 100, targetRect.height / 100)
        case _:
            raise ValueError(f"ROI/DROI '{targetRect.mode}' mode is unsupported")


def scaleDROI(source: Rect, target: DROI) -> Geometry:
    """
    Scale relative DROI to the source image

    Args:
        source: source rect
        target: area of interest
    Returns:
        geometry of interest
    Raises:
        ValueError: If mode value is unsupported
    """
    match target.mode:
        case "abs":
            return target.area
        case "percent":
            return affinity.scale(target.area, xfact=source.width / 100, yfact=source.height / 100, origin=(0, 0))
        case _:
            raise ValueError(f"ROI/DROI '{target.mode}' mode is unsupported")
