from enum import Enum
from typing import Annotated, ClassVar, Generic, Literal

import attr
from annotated_types import Ge, Le
from pydantic import AfterValidator, Field, RootModel, Strict, ValidationError, WrapValidator
from pydantic.root_model import RootModelRootType as TargetRootType
from pydantic_core import PydanticCustomError
from shapely import GeometryCollection, GEOSException, MultiPolygon, Polygon, wkb, wkt
from vlutils.structures.pydantic import BaseModel

UNSET = object()


class TrackStatus(Enum):
    started = "started"
    inProcess = "in_process"
    finished = "finished"


class EventStatus(Enum):
    started = "started"
    inProcess = "in_process"
    finished = "finished"


@attr.dataclass(slots=True)
class VideoSegment:
    """Video segment"""

    startTimeOffset: float
    endTimeOffset: float

    def asDict(self):
        """Build results in API format."""
        return {
            "start_time_offset": self.startTimeOffset,
            "end_time_offset": self.endTimeOffset,
        }


def _customizeUnionError(values, handler):
    """
    WrapValidator function that is applied on Union pydantic fields with discriminator
    to exclude discriminator value from exception location ('loc') on trace,
    and patch errors about discriminator field
    """
    try:
        return handler(values)
    except ValidationError as exc:
        error = exc.errors()[0]
        if error["type"] in ("union_tag_not_found", "union_tag_invalid"):
            error["loc"] = (error["ctx"]["discriminator"].replace("'", ""),)
            match error["type"]:
                case "union_tag_not_found":
                    message = "Field required"
                case "union_tag_invalid":
                    message = f"Input should be one of [{error['ctx']['expected_tags']}]"
            error["type"] = PydanticCustomError("discriminator", message, error.get("ctx"))
        elif error["loc"]:
            error["loc"] = error["loc"][1:]
        raise ValidationError.from_exception_data(exc.title, [error])


WrapUnionError = WrapValidator(_customizeUnionError)


def _customizeModelError(values, handler):
    """
    WrapValidator function that is applied on fields to patch errors about model type
    """
    try:
        return handler(values)
    except ValidationError as exc:
        error = exc.errors()[0]
        if error["type"] == "model_type":
            error["msg"] = error["msg"].partition(" or instance of")[0]
            error["type"] = PydanticCustomError("model_type", error["msg"], error.get("ctx"))
            raise ValidationError.from_exception_data(exc.title, [error])
        elif error["type"] == "model_attributes_type":
            error["msg"] = error["msg"].partition(" or object to extract fields from")[0]
            error["type"] = PydanticCustomError("model_type", error["msg"], error.get("ctx"))
            raise ValidationError.from_exception_data(exc.title, [error])
        raise


WrapModelError = WrapValidator(_customizeModelError)


class CommonTargets(RootModel, Generic[TargetRootType]):
    """Base analytics targets stub."""

    _default: ClassVar[set[TargetRootType]]  # type: ignore[misc]
    root: set[TargetRootType] = Field(default=UNSET, validate_default=False)  # type: ignore[arg-type]

    def model_post_init(self, __context) -> None:
        if self.root is UNSET:
            self.root = self._default

    def __iter__(self):
        return iter(self.root)

    def __contains__(self, key):
        return key in self.root


class RateFrame(BaseModel):
    period: Annotated[int, Ge(1), Strict()] = 10
    unit: Literal["frame"] = "frame"


class RateTime(BaseModel):
    period: Annotated[float, Ge(0), Strict()] = 10
    unit: Literal["second"] = "second"


class FrequencyParams(BaseModel):
    rate: Annotated[RateFrame | RateTime, WrapUnionError, Field(discriminator="unit", default_factory=RateFrame)]


class ROI[TCoordinate, TMode](BaseModel):
    x: TCoordinate
    y: TCoordinate
    height: TCoordinate
    width: TCoordinate
    mode: TMode


class DROI[TArea, TFormat](BaseModel):
    area: TArea
    form: TFormat
    mode: Literal["percent", "abs"]


class Point(BaseModel):
    x: Annotated[float, Ge(0.0), Strict()]
    y: Annotated[float, Ge(0.0), Strict()]


def loadWKT(value: str) -> GeometryCollection:
    """
    Load WTK area
    Args:
        value: string value

    Returns:
        valid wkt area
    """
    try:
        area = wkt.loads(value)
    except GEOSException as exc:
        raise PydanticCustomError("ValidationError", f"Invalid wkt coordinates: {exc.args[0]}") from exc

    if not area.is_valid:
        try:
            area.covers(area.convex_hull.centroid)
        except GEOSException as exc:
            raise PydanticCustomError("ValidationError", f"Invalid wkt coordinates: {exc.args[0]}") from exc
        raise PydanticCustomError("ValidationError", f"Invalid wkt coordinates.")

    return area


def loadDROI(value: DROI):
    """Load & validate DROI params."""

    match value.form:
        case "common":
            area = MultiPolygon([Polygon([(point.x, point.y) for point in poligon]) for poligon in value.area])
        case "wkt":
            area = loadWKT(value.area)
        case "wkb":
            area = wkb.loads(value.area)
        case _:
            raise RuntimeError("Format unsupported")
    if value.mode == "percent" and any(x > 100 for x in area.bounds):
        raise PydanticCustomError("ValidationError", "Invalid relative coordinates")

    value.area = area
    return value


DroiPolicy = Annotated[
    DROI[str, Literal["wkt"]]
    # not in demand at the moment
    # | DROI[bytes, Literal["wkb"]]
    | DROI[Annotated[list[Annotated[list[Point], Field(min_length=3)]], Field(min_length=1)], Literal["common"]],
    WrapUnionError,
    WrapModelError,
    Field(discriminator="form", default_factory=lambda: None),
    AfterValidator(loadDROI),
]

RoiPolicy = Annotated[
    ROI[Annotated[float, Ge(0.0), Le(100.0), Strict()], Literal["percent"]]
    | ROI[Annotated[int, Ge(0), Le(65536), Strict()], Literal["abs"]],
    WrapUnionError,
    Field(discriminator="mode", default_factory=lambda: None),
]


class ImageRetainPolicy(BaseModel):
    quality: Annotated[float, Ge(0.0), Le(1.0), Strict(), AfterValidator(lambda x: round(x * 100))] = 100
    maxSize: Annotated[int, Ge(0), Strict()] = 640
    mimetype: Literal["JPEG", "PNG"] = "JPEG"
    ttl: Literal[-1, 1, 2, 3, 4, 5, 6, 7, 14, 30, 60, 90, 180, 365] = Field(default_factory=lambda: None)


class OnetimeEventPolicyEnd(BaseModel):
    trigger: Literal["end"] = "end"


class OnetimeEventPolicyStart(BaseModel):
    trigger: Literal["start"] = "start"


class PeriodEventPolicy(BaseModel):
    trigger: Literal["period"]
    interval: Annotated[float, Ge(0.0), Strict()]


EventPolicy = Annotated[
    OnetimeEventPolicyStart | OnetimeEventPolicyEnd | PeriodEventPolicy, Field(discriminator="trigger")
]


class CommonParams(BaseModel):
    """Base analytics parameters stub."""

    probeCount: Annotated[int, Ge(0), Strict()] = 3
    roi: RoiPolicy
    droi: DroiPolicy
    imageRetainPolicy: Annotated[ImageRetainPolicy, WrapModelError, Field(default_factory=ImageRetainPolicy)]
    eventPolicy: Annotated[EventPolicy, WrapUnionError, WrapModelError, Field(default_factory=OnetimeEventPolicyEnd)]
